# Copyright 2013 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Abstract solver interface definitions (:class:`Model`, :class:`Variable`,
:class:`Constraint`, :class:`Objective`) intended to be subclassed and
extended for individual solvers.

This module defines the API of optlang objects and indicates which methods need to be implemented in
subclassed solver interfaces.
The classes in this module can be used to construct and modify problems, but no optimizations can be done.
"""
from collections import OrderedDict

try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

import inspect
import logging
import sys
import uuid
import warnings
import sympy

import six

import optlang
from optlang.exceptions import IndicatorConstraintsNotSupported

from optlang.util import parse_expr, expr_to_json, is_numeric, SolverTolerances
from optlang import symbolics

from .container import Container

log = logging.getLogger(__name__)

OPTIMAL = 'optimal'
UNDEFINED = 'undefined'
FEASIBLE = 'feasible'
INFEASIBLE = 'infeasible'
NOFEASIBLE = 'nofeasible'
UNBOUNDED = 'unbounded'
INFEASIBLE_OR_UNBOUNDED = 'infeasible_or_unbounded'
LOADED = 'loaded'
CUTOFF = 'cutoff'
ITERATION_LIMIT = 'iteration_limit'
MEMORY_LIMIT = 'memory_limit'
NODE_LIMIT = 'node_limit'
TIME_LIMIT = 'time_limit'
SOLUTION_LIMIT = 'solution_limit'
INTERRUPTED = 'interrupted'
NUMERIC = 'numeric'
SUBOPTIMAL = 'suboptimal'
INPROGRESS = 'in_progress'
ABORTED = 'aborted'
SPECIAL = 'check_original_solver_status'

statuses = {
    OPTIMAL: "An optimal solution as been found.",
    INFEASIBLE: "The problem has no feasible solutions.",
    UNBOUNDED: "The objective can be optimized infinitely.",
    SPECIAL: "The status returned by the solver could not be interpreted. Please refer to the solver's documentation to find the status.",
    UNDEFINED: "The solver determined that the problem is ill-formed. "
    # TODO Add the rest
}


# noinspection PyShadowingBuiltins
class Variable(symbolics.Symbol):
    """Optimization variables.

    Variable objects are used to represents each variable of the optimization problem. When the optimization is
    performed, the combination of variable values that optimizes the objective function, while not violating any
    constraints will be identified. The type of a variable ('continuous', 'integer' or 'binary') can be set using
    the type keyword of the constructor or it can be changed after initialization by :code:`var.type = 'binary'`.

    The variable class subclasses the :code:`sympy.Symbol` class, which means that symbolic expressions of variables
    can be constructed by using regular python syntax, e.g. :code:`my_expression = 2 * var1 + 3 * var2 ** 2`.
    Expressions like this are used when constructing Constraint and Objective objects.
    Once a problem has been optimized, the primal and dual values of a variable can be accessed from the
    primal and dual attributes, respectively.

    Attributes
    ----------
    name: str
        The variable's name.
    lb: float or None, optional
        The lower bound, if None then -inf.
    ub: float or None, optional
        The upper bound, if None then inf.
    type: str, optional
        The variable type, 'continuous' or 'integer' or 'binary'.
    problem: Model or None, optional
        A reference to the optimization model the variable belongs to.

    Examples
    --------
    >>> Variable('x', lb=-10, ub=10)
    '-10 <= x <= 10'
    """

    @staticmethod
    def __test_valid_lower_bound(type, value, name):
        if not (value is None or is_numeric(value)):
            raise TypeError("Variable bounds must be numeric or None.")
        if value is not None:
            if type == 'integer' and value % 1 != 0.:
                raise ValueError(
                    'The provided lower bound %g cannot be assigned to integer variable %s (%g mod 1 != 0).' % (
                        value, name, value))
        if type == 'binary' and (value is None or value not in (0, 1)):
            raise ValueError(
                'The provided lower bound %s cannot be assigned to binary variable %s.' % (value, name))

    @staticmethod
    def __test_valid_upper_bound(type, value, name):
        if not (value is None or is_numeric(value)):
            raise TypeError("Variable bounds must be numeric or None.")
        if value is not None:
            if type == 'integer' and value % 1 != 0.:
                raise ValueError(
                    'The provided upper bound %s cannot be assigned to integer variable %s (%s mod 1 != 0).' % (
                        value, name, value))
        if type == 'binary' and (value is None or value not in (0, 1)):
            raise ValueError(
                'The provided upper bound %s cannot be assigned to binary variable %s.' % (value, name))

    @staticmethod
    def __validate_variable_name(name):
        if len(name) < 1:
            raise ValueError('Variable name must not be empty string')
        for char in name:
            if char.isspace():
                raise ValueError(
                    'Variable names cannot contain whitespace characters. "%s" contains whitespace character "%s".' % (
                        name, char)
                )

    @classmethod
    def clone(cls, variable, **kwargs):
        """
        Make a copy of another variable. The variable being copied can be of the same type or belong to
        a different solver interface.

        Example
        ----------
        >>> var_copy = Variable.clone(old_var)
        """
        return cls(variable.name, lb=variable.lb, ub=variable.ub, type=variable.type, **kwargs)

    def __init__(self, name, lb=None, ub=None, type="continuous", problem=None, *args, **kwargs):

        # Ensure that name is str and not unicode - some solvers only support string type in Python 2.
        if six.PY2:
            name = str(name)

        self.__validate_variable_name(name)

        self._name = name
        symbolics.Symbol.__init__(self, name, *args, **kwargs)
        self._lb = lb
        self._ub = ub
        if self._lb is None and type == 'binary':
            self._lb = 0.
        if self._ub is None and type == 'binary':
            self._ub = 1.
        self.__test_valid_lower_bound(type, self._lb, name)
        self.__test_valid_upper_bound(type, self._ub, name)
        self.problem = None
        self.type = type
        self.problem = problem

    @property
    def name(self):
        """Name of variable."""
        return self._name

    @name.setter
    def name(self, value):
        self.__validate_variable_name(value)
        old_name = getattr(self, 'name', None)
        self._name = value
        if getattr(self, 'problem', None) is not None and value != old_name:
            self.problem.variables.update_key(old_name)
            self.problem._variables_to_constraints_mapping[value] = self.problem._variables_to_constraints_mapping[old_name]
            del self.problem._variables_to_constraints_mapping[old_name]

    @property
    def lb(self):
        """Lower bound of variable."""
        return self._lb

    @lb.setter
    def lb(self, value):
        if hasattr(self, 'ub') and self.ub is not None and value is not None and value > self.ub:
            raise ValueError(
                'The provided lower bound %g is larger than the upper bound %g of variable %s.' % (
                    value, self.ub, self))
        self.__test_valid_lower_bound(self.type, value, self.name)
        self._lb = value
        if self.problem is not None:
            self.problem._pending_modifications.var_lb.append((self, value))

    @property
    def ub(self):
        """Upper bound of variable."""
        return self._ub

    @ub.setter
    def ub(self, value):
        if hasattr(self, 'lb') and self.lb is not None and value is not None and value < self.lb:
            raise ValueError(
                'The provided upper bound %g is smaller than the lower bound %g of variable %s.' % (
                    value, self.lb, self))
        self.__test_valid_upper_bound(self.type, value, self.name)
        self._ub = value
        if self.problem is not None:
            self.problem._pending_modifications.var_ub.append((self, value))

    def set_bounds(self, lb, ub):
        """
        Change the lower and upper bounds of a variable.
        """
        if lb is not None and ub is not None and lb > ub:
            raise ValueError(
                "The provided lower bound {} is larger than the provided upper bound {}".format(lb, ub)
            )
        self._lb = lb
        self._ub = ub
        if self.problem is not None:
            self.problem._pending_modifications.var_lb.append((self, lb))
            self.problem._pending_modifications.var_ub.append((self, ub))

    @property
    def type(self):
        """Variable type ('either continuous, integer, or binary'.)"""
        return self._type

    @type.setter
    def type(self, value):
        if value == 'continuous':
            self._type = value
        elif value == 'integer':
            self._type = value
            try:
                self.lb = round(self.lb)
            except TypeError:
                pass
            try:
                self.ub = round(self.ub)
            except TypeError:
                pass
        elif value == 'binary':
            self._type = value
            self._lb = 0
            self._ub = 1
        else:
            raise ValueError(
                "'%s' is not a valid variable type. Choose between 'continuous, 'integer', or 'binary'." % value)

    @property
    def primal(self):
        """The primal of variable (None if no solution exists)."""
        if self.problem:
            return self._get_primal()
        else:
            return None

    def _get_primal(self):
        return None

    @property
    def dual(self):
        """The dual of variable (None if no solution exists)."""
        return None

    def __str__(self):
        """Print a string representation of variable.

        Examples
        --------
        >>> Variable('x', lb=-10, ub=10)
        '-10 <= x <= 10'
        """
        if self.lb is not None:
            lb_str = str(self.lb) + " <= "
        else:
            lb_str = ""
        if self.ub is not None:
            ub_str = " <= " + str(self.ub)
        else:
            ub_str = ""
        return ''.join((lb_str, super(Variable, self).__str__(), ub_str))

    def __repr__(self):
        """Does exactly the same as __str__ for now."""
        return self.__str__()

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__ = state

    def __reduce__(self):
        return (type(self), (self.name, self.lb, self.ub, self.type, self.problem))

    def to_json(self):
        """
        Returns a json-compatible object from the Variable that can be saved using the json module.

        Example
        --------
        >>> import json
        >>> with open("path_to_file.json", "w") as outfile:
        >>>     json.dump(var.to_json(), outfile)
        """
        json_obj = {
            "name": self.name,
            "lb": self.lb,
            "ub": self.ub,
            "type": self.type
        }
        return json_obj

    @classmethod
    def from_json(cls, json_obj):
        """
        Constructs a Variable from the provided json-object.

        Example
        --------
        >>> import json
        >>> with open("path_to_file.json") as infile:
        >>>     var = Variable.from_json(json.load(infile))
        """
        return cls(json_obj["name"], lb=json_obj["lb"], ub=json_obj["ub"], type=json_obj["type"])

    # def _round_primal_to_bounds(self, primal, tolerance=1e-5):
    #     """Rounds primal value to lie within variables bounds.
    #
    #     Raises if exceeding threshold.
    #
    #     Parameters
    #     ----------
    #     primal : float
    #         The primal value to round.
    #     tolerance : float (optional)
    #         The tolerance threshold (default: 1e-5).
    #     """
    #     if (self.lb is None or primal >= self.lb) and (self.ub is None or primal <= self.ub):
    #         return primal
    #     else:
    #         if (primal <= self.lb) and ((self.lb - primal) <= tolerance):
    #             return self.lb
    #         elif (primal >= self.ub) and ((self.ub - primal) >= -tolerance):
    #             return self.ub
    #         else:
    #             raise AssertionError(
    #                 'The primal value %s returned by the solver is out of bounds for variable %s (lb=%s, ub=%s)' % (
    #                     primal, self.name, self.lb, self.ub))


# noinspection PyPep8Naming
class OptimizationExpression(object):
    """Abstract base class for Objective and Constraint."""

    @staticmethod
    def _validate_optimization_expression_name(name):
        if name is None:
            return
        if len(name) < 1:
            raise ValueError('Name must not be empty string')
        for char in name:
            if char.isspace():
                raise ValueError(
                    'Names cannot contain whitespace characters. "%s" contains whitespace character "%s".' % (
                        name, char)
                )

    @classmethod
    def _substitute_variables(cls, expression, model=None, **kwargs):
        """Substitutes variables in (optimization)expression (constraint/objective) with variables of the appropriate interface type.
        Attributes
        ----------
        expression: Constraint, Objective
            An optimization expression.
        model: Model or None, optional
            A reference to an optimization model that should be searched for appropriate variables first.
        """
        interface = sys.modules[cls.__module__]
        variable_substitutions = dict()
        for variable in expression.variables:
            if model is not None and variable.name in model.variables:
                # print(variable.name, id(variable.problem))
                variable_substitutions[variable] = model.variables[variable.name]
            else:
                variable_substitutions[variable] = interface.Variable.clone(variable)
        adjusted_expression = expression.expression.xreplace(variable_substitutions)
        return adjusted_expression

    def __init__(self, expression, name=None, problem=None, sloppy=False, *args, **kwargs):
        # Ensure that name is str and not unicode - some solvers only support string type in Python 2.
        if six.PY2 and name is not None:
            name = str(name)

        self._validate_optimization_expression_name(name)

        super(OptimizationExpression, self).__init__(*args, **kwargs)
        self._problem = problem
        if sloppy:
            self._expression = expression
        else:
            self._expression = self._canonicalize(expression)
        if name is None:
            self._name = str(uuid.uuid1())
        else:
            self._name = name

    @property
    def name(self):
        """The name of the object"""
        return self._name

    @name.setter
    def name(self, value):
        self._validate_optimization_expression_name(value)
        self._name = value

    @property
    def problem(self):
        """A reference to the model that the object belongs to (or None)"""
        return getattr(self, '_problem', None)

    @problem.setter
    def problem(self, value):
        self._problem = value

    def _get_expression(self):
        return self._expression

    @property
    def expression(self):
        """The mathematical expression defining the objective/constraint."""
        return self._get_expression()

    @property
    def variables(self):
        """Variables in constraint/objective's expression."""
        return self.expression.atoms(Variable)

    def _canonicalize(self, expression):
        if isinstance(expression, float):
            return symbolics.Real(expression)
        elif isinstance(expression, int):
            return symbolics.Integer(expression)
        else:
            if optlang._USING_SYMENGINE:
                expression = expression.expand()  # This is a good way to canonicalize, but is quite slow for sympy
            return expression

    @property
    def is_Linear(self):
        """Returns True if expression is linear (a polynomial with degree 1 or 0) (read-only)."""
        coeff_dict = self.expression.as_coefficients_dict()
        for key in coeff_dict.keys():
            if len(key.free_symbols) < 2 and (key.is_Add or key.is_Mul or key.is_Atom):
                pass
            else:
                return False
            if key.is_Pow and key.args[1] != 1:
                return False
        else:
            return True

    @property
    def is_Quadratic(self):
        """Returns True if the expression is a polynomial with degree exactly 2 (read-only)."""
        if self.expression.is_Atom:
            return False
        if all((len(key.free_symbols) < 2 and (key.is_Add or key.is_Mul or key.is_Atom)
                for key in self.expression.as_coefficients_dict().keys())):
            return False
        if self.expression.is_Add:
            terms = self.expression.args
            is_quad = False
            for term in terms:
                if len(term.free_symbols) > 2:
                    return False
                if term.is_Pow:
                    if not term.args[1].is_Number or term.args[1] > 2:
                        return False
                    else:
                        is_quad = True
                elif term.is_Mul:
                    if len(term.free_symbols) == 2:
                        is_quad = True
                    if term.args[1].is_Pow:
                        if not term.args[1].args[1].is_Number or term.args[1].args[1] > 2:
                            return False
                        else:
                            is_quad = True
            return is_quad
        else:
            if isinstance(self.expression, sympy.Basic):
                sympy_expression = self.expression
            else:
                sympy_expression = sympy.sympify(self.expression)
            # TODO: Find a way to do this with symengine (Poly is not part of symengine, 23 March 2017)
            poly = sympy_expression.as_poly(*sympy_expression.atoms(sympy.Symbol))
            if poly is None:
                return False
            else:
                return poly.is_quadratic

    def __iadd__(self, other):
        self._expression += other
        return self

    def __isub__(self, other):
        self._expression -= other
        return self

    def __imul__(self, other):
        self._expression *= other
        return self

    def __idiv__(self, other):
        self._expression /= other
        return self

    def __itruediv__(self, other):
        self._expression /= other
        return self

    def set_linear_coefficients(self, coefficients):
        """Set coefficients of linear terms in constraint or objective.
        Existing coefficients for linear or non-linear terms will not be modified.

        Note: This method interacts with the low-level solver backend and can only be used on objects that are
        associated with a Model. The method is not part of optlangs basic interface and should be used mainly where
        speed is important.

        Parameters
        ----------
        coefficients : dict
            A dictionary like {variable1: coefficient1, variable2: coefficient2, ...}

        Returns
        -------
        None
        """
        raise NotImplementedError("Child classes should implement this.")

    def get_linear_coefficients(self, variables):
        """Get coefficients of linear terms in constraint or objective.

        Note: This method interacts with the low-level solver backend and can only be used on objects that are
        associated with a Model. The method is not part of optlangs basic interface and should be used mainly where
        speed is important.

        Parameters
        ----------
        variables : iterable
            An iterable of Variable objects

        Returns
        -------
        Coefficients : dict
            {var1: coefficient, var2: coefficient ...}
        """
        raise NotImplementedError("Child classes should implement this.")


class Constraint(OptimizationExpression):
    """
    Constraint objects represent the mathematical (in-)equalities that constrain an optimization problem.
    A constraint is formulated by a symbolic expression of variables and a lower and/or upper bound.
    Equality constraints can be formulated by setting the upper and lower bounds to the same value.

    Some solvers support indicator variables. This lets a binary variable act as a switch that decides whether
    the constraint should be active (cannot be violated) or inactive (can be violated).

    The constraint expression can be an arbitrary combination of variables, however the individual solvers
    have limits to the forms of constraints they allow. Most solvers only allow linear constraints, meaning that
    the expression should be of the form :code:`a * var1 + b * var2 + c * var3 ...`

    Attributes
    ----------
    expression: sympy
        The mathematical expression defining the constraint.
    name: str, optional
        The constraint's name.
    lb: float or None, optional
        The lower bound, if None then -inf.
    ub: float or None, optional
        The upper bound, if None then inf.
    indicator_variable: Variable
        The indicator variable (needs to be binary).
    active_when: 0 or 1 (default 0)
        When the constraint should
    problem: Model or None, optional
        A reference to the optimization model the variable belongs to.

    Examples
    ----------
    >>> expr = 2.4 * var1 - 3.8 * var2
    >>> c1 = Constraint(expr, lb=0, ub=10)

    >>> indicator_var = Variable("var3", type="binary") # Only possible with some solvers
    >>> c2 = Constraint(var2, lb=0, ub=0, indicator_variable=indicator_var, active_when=1) # When the indicator is 1, var2 is constrained to be 0
    """


    _INDICATOR_CONSTRAINT_SUPPORT = True

    def _check_valid_lower_bound(self, value):
        if not (value is None or is_numeric(value)):
            raise TypeError("Constraint bounds must be numeric or None, not {}".format(type(value)))
        if value is not None and getattr(self, "ub", None) is not None and value > self.ub:
            raise ValueError("Cannot set a lower bound that is greater than the upper bound.")

    def _check_valid_upper_bound(self, value):
        if not (value is None or is_numeric(value)):
            raise TypeError("Constraint bounds must be numeric or None, not {}".format(type(value)))
        if value is not None and getattr(self, "lb", None) is not None and value < self.lb:
            raise ValueError("Cannot set an upper bound that is less than the lower bound.")

    @classmethod
    def __check_valid_indicator_variable(cls, variable):
        if variable is not None and not cls._INDICATOR_CONSTRAINT_SUPPORT:
            raise IndicatorConstraintsNotSupported(
                'Solver interface %s does not support indicator constraints' % cls.__module__)
        if variable is not None and variable.type != 'binary':
            raise ValueError('Provided indicator variable %s is not binary.' % variable)

    @staticmethod
    def __check_valid_active_when(active_when):
        if active_when != 0 and active_when != 1:
            raise ValueError('Provided active_when argument %s needs to be either 1 or 0' % active_when)

    @classmethod
    def clone(cls, constraint, model=None, **kwargs):
        """
        Make a copy of another constraint. The constraint being copied can be of the same type or belong to
        a different solver interface.

        Parameters
        ----------
        constraint: interface.Constraint (or subclass)
            The constraint to copy
        model: Model or None
            The variables of the new constraint will be taken from this model. If None, new variables will be
            constructed.

        Example
        ----------
        >>> const_copy = Constraint.clone(old_constraint)
        """
        return cls(cls._substitute_variables(constraint, model=model), lb=constraint.lb, ub=constraint.ub,
                   indicator_variable=constraint.indicator_variable, active_when=constraint.active_when,
                   name=constraint.name, sloppy=True, **kwargs)

    def __init__(self, expression, lb=None, ub=None, indicator_variable=None, active_when=1, *args, **kwargs):
        self._problem = None
        self.lb = lb
        self.ub = ub
        super(Constraint, self).__init__(expression, *args, **kwargs)
        self.__check_valid_indicator_variable(indicator_variable)
        self.__check_valid_active_when(active_when)
        self._indicator_variable = indicator_variable
        self._active_when = active_when

    @property
    def lb(self):
        """Lower bound of constraint."""
        return self._lb

    @lb.setter
    def lb(self, value):
        self._check_valid_lower_bound(value)
        self._lb = value

    @property
    def ub(self):
        """Upper bound of constraint."""
        return self._ub

    @ub.setter
    def ub(self, value):
        self._check_valid_upper_bound(value)
        self._ub = value

    @OptimizationExpression.name.setter
    def name(self, value):
        old_name = getattr(self, 'name', None)
        super(Constraint, Constraint).name.fset(self, value)
        if getattr(self, 'problem', None) is not None and value != old_name:
            self.problem.constraints.update_key(old_name)

    @property
    def indicator_variable(self):
        """The indicator variable of constraint (if available)."""
        return self._indicator_variable

    # @indicator_variable.setter
    # def indicator_variable(self, value):
    #     self.__check_valid_indicator_variable(value)
    #     self._indicator_variable = value

    @property
    def active_when(self):
        """Activity relation of constraint to indicator variable (if supported)."""
        return self._active_when

    def __str__(self):
        if self.lb is not None:
            lhs = str(self.lb) + ' <= '
        else:
            lhs = ''
        if self.ub is not None:
            rhs = ' <= ' + str(self.ub)
        else:
            rhs = ''
        if self.indicator_variable is not None:
            lhs = self.indicator_variable.name + ' = ' + str(self.active_when) + ' -> ' + lhs
        return str(self.name) + ": " + lhs + self.expression.__str__() + rhs

    def _canonicalize(self, expression):
        expression = super(Constraint, self)._canonicalize(expression)
        if expression.is_Atom or expression.is_Mul:
            return expression
        lonely_coeffs = [arg for arg in expression.args if arg.is_Number]
        if not lonely_coeffs:
            return expression
        assert len(lonely_coeffs) == 1
        coeff = lonely_coeffs[0]
        expression = expression - coeff
        if self.lb is not None and self.ub is not None:
            oldub = self.ub
            self.ub = None
            self.lb = self.lb - float(coeff)
            self.ub = oldub - float(coeff)
        elif self.lb is not None:
            self.lb = self.lb - float(coeff)
        elif self.ub is not None:
            self.ub = self.ub - float(coeff)
        else:
            raise ValueError(
                "%s cannot be shaped into canonical form if neither lower or upper constraint bounds are set."
                % expression
            )
        return expression

    @property
    def primal(self):
        """Primal of constraint (None if no solution exists)."""
        return None

    @property
    def dual(self):
        """Dual of constraint (None if no solution exists)."""
        return None

    def _round_primal_to_bounds(self, primal, tolerance=1e-5):
        if (self.lb is None or primal >= self.lb) and (self.ub is None or primal <= self.ub):
            return primal
        else:
            if (primal <= self.lb) and ((self.lb - primal) <= tolerance):
                return self.lb
            elif (primal >= self.ub) and ((self.ub - primal) >= -tolerance):
                return self.ub
            else:
                raise AssertionError(
                    'The primal value %s returned by the solver is out of bounds for variable %s (lb=%s, ub=%s)' % (
                        primal, self.name, self.lb, self.ub))

    def to_json(self):
        """
        Returns a json-compatible object from the constraint that can be saved using the json module.

        Example
        --------
        >>> import json
        >>> with open("path_to_file.json", "w") as outfile:
        >>>     json.dump(constraint.to_json(), outfile)
        """
        if self.indicator_variable is None:
            indicator = None
        else:
            indicator = self.indicator_variable.name
        json_obj = {
            "name": self.name,
            "expression": expr_to_json(self.expression),
            "lb": self.lb,
            "ub": self.ub,
            "indicator_variable": indicator,
            "active_when": self.active_when
        }
        return json_obj

    @classmethod
    def from_json(cls, json_obj, variables=None):
        """
        Constructs a Variable from the provided json-object.

        Example
        --------
        >>> import json
        >>> with open("path_to_file.json") as infile:
        >>>     constraint = Constraint.from_json(json.load(infile))
        """
        if variables is None:
            variables = {}
        expression = parse_expr(json_obj["expression"], variables)
        if json_obj["indicator_variable"] is None:
            indicator = None
        else:
            indicator = variables[json_obj["indicator_variable"]]
        return cls(
            expression,
            name=json_obj["name"],
            lb=json_obj["lb"],
            ub=json_obj["ub"],
            indicator_variable=indicator,
            active_when=json_obj["active_when"]
        )


class Objective(OptimizationExpression):
    """
    Objective objects are used to represent the objective function of an optimization problem.
    An objective consists of a symbolic expression of variables in the problem and a direction. The direction
    can be either 'min' or 'max' and specifies whether the problem is a minimization or a maximization problem.

    After a problem has been optimized, the optimal objective value can be accessed from the 'value' attribute
    of the model's objective, i.e. :code:`obj_val = model.objective.value`.

    Attributes
    ----------
    expression: sympy
        The mathematical expression defining the objective.
    name: str, optional
        The name of the constraint.
    direction: 'max' or 'min'
        The optimization direction.
    value: float, read-only
        The current objective value.
    problem: solver
        The low-level solver object.

    """

    @classmethod
    def clone(cls, objective, model=None, **kwargs):
        """
        Make a copy of an objective. The objective being copied can be of the same type or belong to
        a different solver interface.

        Example
        ----------
        >>> new_objective = Objective.clone(old_objective)
        """
        return cls(cls._substitute_variables(objective, model=model), name=objective.name,
                   direction=objective.direction, sloppy=True, **kwargs)

    def __init__(self, expression, value=None, direction='max', *args, **kwargs):
        self._value = value
        self._direction = direction
        super(Objective, self).__init__(expression, *args, **kwargs)

    @property
    def value(self):
        """The objective value."""
        return self._value

    def __str__(self):
        return {'max': 'Maximize', 'min': 'Minimize'}[self.direction] + '\n' + str(self.expression)
        # return ' '.join((self.direction, str(self.expression)))

    def __eq__(self, other):
        """Tests *mathematical* equality for two Objectives. Solver specific type does NOT have to match.
        Expression and direction must be the same.
        Name does not have to match"""
        if isinstance(other, Objective):
            return self.expression == other.expression and self.direction == other.direction
        else:
            return False
            #

    def _canonicalize(self, expression):
        """For example, changes x + y to 1.*x + 1.*y"""
        expression = super(Objective, self)._canonicalize(expression)
        if isinstance(expression, sympy.Basic):
            expression *= 1.
        else:  # pragma: no cover   # symengine
            expression = (1. * expression).expand()
        return expression

    @property
    def direction(self):
        """The direction of optimization. Either 'min' or 'max'."""
        return self._direction

    @direction.setter
    def direction(self, value):
        if value not in ['max', 'min']:
            raise ValueError("Provided optimization direction %s is neither 'min' or 'max'." % value)
        self._direction = value

    def set_linear_coefficients(self, coefficients):
        """Set linear coefficients in objective.

        coefficients : dict
            A dictionary of the form {variable1: coefficient1, variable2: coefficient2, ...}
        """
        raise NotImplementedError("Child class should implement this.")

    def to_json(self):
        """
        Returns a json-compatible object from the objective that can be saved using the json module.

        Example
        --------
        >>> import json
        >>> with open("path_to_file.json", "w") as outfile:
        >>>     json.dump(obj.to_json(), outfile)
        """
        json_obj = {
            "name": self.name,
            "expression": expr_to_json(self.expression),
            "direction": self.direction
        }
        return json_obj

    @classmethod
    def from_json(cls, json_obj, variables=None):
        """
        Constructs an Objective from the provided json-object.

        Example
        --------
        >>> import json
        >>> with open("path_to_file.json") as infile:
        >>>     obj = Objective.from_json(json.load(infile))
        """
        if variables is None:
            variables = {}
        expression = parse_expr(json_obj["expression"], variables)
        return cls(
            expression,
            direction=json_obj["direction"],
            name=json_obj["name"]
        )


class Configuration(object):
    """
    Optimization solver configuration.
    This object allows the user to change certain parameters and settings in the solver.
    It is meant to allow easy access to a few common and important parameters. For information on changing
    other solver parameters, please consult the documentation from the solver provider.
    Some changeable parameters are listed below. Note that some solvers might not implement all of these
    and might also implement additional parameters.

    Attributes
    ----------
    verbosity: int from 0 to 3
        Changes the level of output.
    timeout: int or None
        The time limit in second the solver will use to optimize the problem.
    presolve: Boolean or 'auto'
        Tells the solver whether to use (solver-specific) pre-processing to simplify the problem.
        This can decrease solution time, but also introduces overhead. If set to 'auto' the solver will
        first try to solve without pre-processing, and only turn in on in case no optimal solution can be found.
    lp_method: str
        Select which algorithm the LP solver uses, e.g. simplex, barrier, etc.

    """

    @classmethod
    def clone(cls, config, problem=None, **kwargs):
        properties = (k for k, v in inspect.getmembers(cls, predicate=inspect.isdatadescriptor) if
                      not k.startswith('__'))
        parameters = {property: getattr(config, property) for property in properties if hasattr(config, property)}
        if hasattr(config, "tolerances"):
            parameters["tolerances"] = config.tolerances.to_dict()
        return cls(problem=problem, **parameters)

    def __init__(self, problem=None, *args, **kwargs):
        self.problem = problem
        self._add_tolerances()

    @property
    def verbosity(self):
        """Verbosity level.

        0: no output
        1: error and warning messages only
        2: normal output
        3: full output
        """
        raise NotImplementedError

    @verbosity.setter
    def verbosity(self, value):
        raise NotImplementedError

    @property
    def timeout(self):
        """Timeout parameter (seconds)."""
        raise NotImplementedError

    @timeout.setter
    def timeout(self):
        raise NotImplementedError

    @property
    def presolve(self):
        """
        Turn pre-processing on or off. Set to 'auto' to only use presolve if no optimal solution can be found.
        """
        raise NotImplementedError

    @presolve.setter
    def presolve(self):
        raise NotImplementedError

    def _add_tolerances(self):
        self.tolerances = SolverTolerances(self._tolerance_functions())

    def _tolerance_functions(self):
        """
        This should be implemented in child classes. Must return a dict, where keys are available tolerance parameters
        and values are tuples of (getter_function, setter_function).
        The getter functions must be callable with no arguments and the setter functions must be callable with the
        new value as the only argument
        """
        return {}

    def __setstate__(self, state):
        self.__init__()


class MathematicalProgrammingConfiguration(Configuration):
    def __init__(self, *args, **kwargs):
        super(MathematicalProgrammingConfiguration, self).__init__(*args, **kwargs)

    @property
    def presolve(self):
        """If the presolver should be used (if available)."""
        raise NotImplementedError

    @presolve.setter
    def presolve(self, value):
        raise NotImplementedError


class EvolutionaryOptimizationConfiguration(Configuration):
    """docstring for HeuristicOptimization"""

    def __init__(self, *args, **kwargs):
        super(EvolutionaryOptimizationConfiguration, self).__init__(*args, **kwargs)


class Model(object):
    """
    The model object represents an optimization problem and contains the variables, constraints an objective that
    make up the problem. Variables and constraints can be added and removed using the :code:`.add` and :code:`.remove` methods,
    while the objective can be changed by setting the objective attribute,
    e.g. :code:`model.objective = Objective(expr, direction="max")`.

    Once the problem has been formulated the optimization can be performed by calling the :code:`.optimize` method.
    This will return the status of the optimization, most commonly 'optimal', 'infeasible' or 'unbounded'.

    Attributes
    ----------
    objective: str
        The objective function.
    name: str, optional
        The name of the optimization problem.
    variables: Container, read-only
        Contains the variables of the optimization problem.
        The keys are the variable names and values are the actual variables.
    constraints: Container, read-only
         Contains the variables of the optimization problem.
         The keys are the constraint names and values are the actual constraints.
    status: str, read-only
        The status of the optimization problem.

    Examples
    --------
    >>> model = Model(name="my_model")
    >>> x1 = Variable("x1", lb=0, ub=20)
    >>> x2 = Variable("x2", lb=0, ub=10)
    >>> c1 = Constraint(2 * x1 - x2, lb=0, ub=0) # Equality constraint
    >>> model.add([x1, x2, c1])
    >>> model.objective = Objective(x1 + x2, direction="max")
    >>> model.optimize()
    'optimal'
    >>> x1.primal, x2.primal
    '(5.0, 10.0)'

    """

    @classmethod
    def clone(cls, model, use_json=True, use_lp=False):
        """
        Make a copy of a model. The model being copied can be of the same type or belong to
        a different solver interface. This is the preferred way of copying models.

        Example
        ----------
        >>> new_model = Model.clone(old_model)
        """
        model.update()
        interface = sys.modules[cls.__module__]

        if use_lp:
            warnings.warn("Cloning with LP formats can change variable and constraint ID's.")
            new_model = cls.from_lp(model.to_lp())
            new_model.configuration = interface.Configuration.clone(model.configuration, problem=new_model)
            return new_model

        if use_json:
            new_model = cls.from_json(model.to_json())
            new_model.configuration = interface.Configuration.clone(model.configuration, problem=new_model)
            return new_model

        new_model = cls()
        for variable in model.variables:
            new_variable = interface.Variable.clone(variable)
            new_model._add_variable(new_variable)
        for constraint in model.constraints:
            new_constraint = interface.Constraint.clone(constraint, model=new_model)
            new_model._add_constraint(new_constraint)
        if model.objective is not None:
            new_model.objective = interface.Objective.clone(model.objective, model=new_model)
        new_model.configuration = interface.Configuration.clone(model.configuration, problem=new_model)
        return new_model

    def __init__(self, name=None, objective=None, variables=None, constraints=None, problem=None, *args, **kwargs):
        super(Model, self).__init__(*args, **kwargs)

        self._objective = self.interface.Objective(0)
        self._objective.problem = self
        self._variables = Container()
        self._constraints = Container()
        self._variables_to_constraints_mapping = dict()
        self._status = None
        self.name = name

        class Modifications():

            def __init__(self):
                self.add_var = []
                self.add_constr = []
                self.add_constr_sloppy = []
                self.rm_var = []
                self.rm_constr = []
                self.var_lb = []
                self.var_ub = []
                self.toggle = 'add'

            def __str__(self):
                return str(self.__dict__)

        self._pending_modifications = Modifications()

        if problem is not None:
            if not (objective is None and variables is None and constraints is None):
                raise ValueError("Models constructor must be called with the problem argument OR with variables, constraints and objective")
            self._initialize_model_from_problem(problem)

        else:
            self._initialize_problem()
            if variables is not None:
                self.add(variables)
            if constraints is not None:
                self.add(constraints)
            if objective is not None:
                self.objective = objective
        self._initialize_configuration()

    def _initialize_problem(self):
        """
        Constructs an empty solver-specific problem. Is called during construction of new Model objects when problem=None.
        Should be implemented in all solver interfaces.
        """
        pass

    def _initialize_model_from_problem(self, problem):
        """
        Parses a solver-specific problem object and initializes the Model object to a state that is consistent with the problem.
        The problem is set as the Model's corresponding problem.
        Should be implemented in all solver interfaces.
        """
        raise NotImplementedError

    def _initialize_configuration(self):
        """
        Initializes a Configuration object. Should be implemented in solver interfaces if non-standard behaviour is needed
        """
        self.configuration = self.interface.Configuration(problem=self)

    @property
    def interface(self):
        """Provides access to the solver interface the model belongs to

        Returns a Python module, for example optlang.glpk_interface
        """
        return sys.modules[self.__module__]

    @property
    def objective(self):
        """The model's objective function."""
        return self._objective

    @objective.setter
    def objective(self, value):
        self.update()
        for atom in sorted(value.expression.atoms(Variable), key=lambda v: v.name):
            if isinstance(atom, Variable) and (atom.problem is None or atom.problem != self):
                self._pending_modifications.add_var.append(atom)
        self.update()
        if self._objective is not None:
            self._objective.problem = None
        self._objective = value
        self._objective.problem = self

    @property
    def variables(self):
        """The model variables."""
        self.update()
        return self._variables

    @property
    def constraints(self):
        """The model constraints."""
        self.update()
        return self._constraints

    @property
    def status(self):
        """The solver status of the model."""
        return self._status

    def _get_variables_names(self):
        """The names of model variables.

        Returns
        -------
        list
        """
        return [variable.name for variable in self.variables]

    @property
    def primal_values(self):
        """The primal values of model variables.

        The primal values are rounded to the bounds.
        Returns
        -------
        collections.OrderedDict
        """
        return OrderedDict(
            zip(self._get_variables_names(), self._get_primal_values())
        )

    def _get_primal_values(self):
        """The primal values of model variables.

        Returns
        -------
        list
        """
        # Fallback, if nothing faster is available
        return [variable.primal for variable in self.variables]

    @property
    def reduced_costs(self):
        """The reduced costs/dual values of all variables.

        Returns
        -------
        collections.OrderedDict
        """
        return OrderedDict(
            zip(self._get_variables_names(), self._get_reduced_costs())
        )

    def _get_reduced_costs(self):
        """The reduced costs/dual values of all variables.

        Returns
        -------
        list
        """
        # Fallback, if nothing faster is available
        return [variable.dual for variable in self.variables]

    def _get_constraint_names(self):
        """The names of model constraints.

        Returns
        -------
        list
        """
        return [constraint.name for constraint in self.constraints]

    @property
    def constraint_values(self):
        """The primal values of all constraints.

        Returns
        -------
        collections.OrderedDict
        """
        return OrderedDict(
            zip(self._get_constraint_names(), self._get_constraint_values())
        )

    def _get_constraint_values(self):
        """The primal values of all constraints.

        Returns
        -------
        list
        """
        # Fallback, if nothing faster is available
        return [constraint.primal for constraint in self.constraints]

    @property
    def shadow_prices(self):
        """The shadow prices of model (dual values of all constraints).

        Returns
        -------
        collections.OrderedDict
        """
        return OrderedDict(
            zip(self._get_constraint_names(), self._get_shadow_prices())
        )

    def _get_shadow_prices(self):
        """The shadow prices of model (dual values of all constraints).

        Returns
        -------
        collections.OrderedDict
        """
        # Fallback, if nothing faster is available
        return [constraint.dual for constraint in self.constraints]

    @property
    def is_integer(self):
        return any(var.type in ("integer", "binary") for var in self.variables)

    def __str__(self):  # pragma: no cover
        if hasattr(self, "to_lp"):
            return self.to_lp()
        self.update()
        return '\n'.join((
            str(self.objective),
            "subject to",
            '\n'.join([str(constr) for constr in self.constraints]),
            'Bounds',
            '\n'.join([str(var) for var in self.variables])
        ))

    def add(self, stuff, sloppy=False):
        """Add variables and constraints.

        Parameters
        ----------
        stuff : iterable, Variable, Constraint
            Either an iterable containing variables and constraints or a single variable or constraint.

        sloppy : bool
            Check constraints for variables that are not part of the model yet.

        Returns
        -------
        None
        """
        if self._pending_modifications.toggle == 'remove':
            self.update()
            self._pending_modifications.toggle = 'add'
        if isinstance(stuff, Iterable):
            for elem in stuff:
                self.add(elem, sloppy=sloppy)
        elif isinstance(stuff, Variable):
            if stuff.__module__ != self.__module__:
                raise TypeError("Cannot add Variable %s of interface type %s to model of type %s." % (
                    stuff, stuff.__module__, self.__module__))
            self._pending_modifications.add_var.append(stuff)
        elif isinstance(stuff, Constraint):
            if stuff.__module__ != self.__module__:
                raise TypeError("Cannot add Constraint %s of interface type %s to model of type %s." % (
                    stuff, stuff.__module__, self.__module__))
            if sloppy is True:
                self._pending_modifications.add_constr_sloppy.append(stuff)
            else:
                self._pending_modifications.add_constr.append(stuff)
        else:
            raise TypeError("Cannot add %s. It is neither a Variable, or Constraint." % stuff)

    def remove(self, stuff):
        """Remove variables and constraints.

        Parameters
        ----------
        stuff : iterable, str, Variable, Constraint
            Either an iterable containing variables and constraints to be removed from the model or a single variable or contstraint (or their names).

        Returns
        -------
        None
        """
        if self._pending_modifications.toggle == 'add':
            self.update()
            self._pending_modifications.toggle = 'remove'
        if isinstance(stuff, str):
            try:
                variable = self.variables[stuff]
                self._pending_modifications.rm_var.append(variable)
            except KeyError:
                try:
                    constraint = self.constraints[stuff]
                    self._pending_modifications.rm_constr.append(constraint)
                except KeyError:
                    raise LookupError(
                        "%s is neither a variable nor a constraint in the current solver instance." % stuff)
        elif isinstance(stuff, Variable):
            self._pending_modifications.rm_var.append(stuff)
        elif isinstance(stuff, Constraint):
            self._pending_modifications.rm_constr.append(stuff)
        elif isinstance(stuff, Iterable):
            for elem in stuff:
                self.remove(elem)
        elif isinstance(stuff, Objective):
            raise TypeError(
                "Cannot remove objective %s. Use model.objective = Objective(...) to change the current objective." % stuff)
        else:
            raise TypeError(
                "Cannot remove %s. It neither a variable or constraint." % stuff)

    def update(self, callback=int):
        """Process all pending model modifications."""
        # print(self._pending_modifications)
        add_var = self._pending_modifications.add_var
        if len(add_var) > 0:
            self._add_variables(add_var)
            self._pending_modifications.add_var = []
        callback()

        add_constr = self._pending_modifications.add_constr
        if len(add_constr) > 0:
            self._add_constraints(add_constr)
            self._pending_modifications.add_constr = []

        add_constr_sloppy = self._pending_modifications.add_constr_sloppy
        if len(add_constr_sloppy) > 0:
            self._add_constraints(add_constr_sloppy, sloppy=True)
            self._pending_modifications.add_constr_sloppy = []

        var_lb = self._pending_modifications.var_lb
        var_ub = self._pending_modifications.var_ub
        if len(var_lb) > 0 or len(var_ub) > 0:
            self._set_variable_bounds_on_problem(var_lb, var_ub)
            self._pending_modifications.var_lb = []
            self._pending_modifications.var_ub = []

        rm_var = self._pending_modifications.rm_var
        if len(rm_var) > 0:
            self._remove_variables(rm_var)
            self._pending_modifications.rm_var = []
        callback()

        rm_constr = self._pending_modifications.rm_constr
        if len(rm_constr) > 0:
            self._remove_constraints(rm_constr)
            self._pending_modifications.rm_constr = []

    def optimize(self):
        """
        Solve the optimization problem using the relevant solver back-end.
        The status returned by this method tells whether an optimal solution was found,
        if the problem is infeasible etc. Consult optlang.statuses for more elaborate explanations
        of each status.

        The objective value can be accessed from 'model.objective.value', while the solution can be
        retrieved by 'model.primal_values'.

        Returns
        -------
        status: str
            Solution status.
        """
        self.update()
        status = self._optimize()
        if status != OPTIMAL and self.configuration.presolve == "auto":
            self.configuration.presolve = True
            status = self._optimize()
            self.configuration.presolve = "auto"
        self._status = status
        return status

    def _optimize(self):
        raise NotImplementedError(
            "You're using the high level interface to optlang. Problems cannot be optimized in this mode. Choose from one of the solver specific interfaces.")

    def _set_variable_bounds_on_problem(self, var_lb, var_ub):
        """"""
        pass

    def _add_variable(self, variable):
        self._add_variables([variable])

    def _add_variables(self, variables):
        for variable in variables:
            self._variables.append(variable)
            self._variables_to_constraints_mapping[variable.name] = set([])
            variable.problem = self

    def _remove_variables(self, variables):
        for variable in variables:
            try:
                self._variables[variable.name]
            except KeyError:
                raise LookupError("Variable %s not in solver" % variable.name)

        constraint_ids = set()
        for variable in variables:
            constraint_ids.update(self._variables_to_constraints_mapping[variable.name])
            del self._variables_to_constraints_mapping[variable.name]
            variable.problem = None
            del self._variables[variable.name]

        replacements = dict([(variable, 0) for variable in variables])
        for constraint_id in constraint_ids:
            constraint = self._constraints[constraint_id]
            constraint._expression = constraint._expression.xreplace(replacements)
        if self.objective is not None:
            self.objective._expression = self.objective._expression.xreplace(replacements)

    def _remove_variable(self, variable):
        self._remove_variables([variable])

    def _add_constraint(self, constraint, sloppy=False):
        self._add_constraints([constraint], sloppy=sloppy)

    def _add_constraints(self, constraints, sloppy=False):
        for constraint in constraints:
            constraint_id = constraint.name
            if sloppy is False:
                variables = constraint.variables
                if constraint.indicator_variable is not None:
                    variables.add(constraint.indicator_variable)
                missing_vars = [var for var in variables if var.problem is not self]
                if len(missing_vars) > 0:
                    self._add_variables(missing_vars)
                for var in variables:
                    try:
                        self._variables_to_constraints_mapping[var.name].add(constraint_id)
                    except KeyError:
                        self._variables_to_constraints_mapping[var.name] = set([constraint_id])
            self._constraints.append(constraint)
            constraint._problem = self

    def _remove_constraints(self, constraints):
        keys = [constraint.name for constraint in constraints]
        if len(constraints) > 350:  # Need to figure out a good threshold here
            self._constraints = self._constraints.fromkeys(set(self._constraints.keys()).difference(set(keys)))
        else:
            for constraint in constraints:
                try:
                    del self._constraints[constraint.name]
                except KeyError:
                    raise LookupError("Constraint %s not in solver" % constraint)
                else:
                    constraint.problem = None

    def _remove_constraint(self, constraint):
        self._remove_constraints([constraint])

    def to_json(self):
        """
        Returns a json-compatible object from the model that can be saved using the json module.
        Variables, constraints and objective contained in the model will be saved. Configurations
        will not be saved.

        Example
        --------
        >>> import json
        >>> with open("path_to_file.json", "w") as outfile:
        >>>     json.dump(model.to_json(), outfile)
        """
        json_obj = {
            "name": self.name,
            "variables": [var.to_json() for var in self.variables],
            "constraints": [const.to_json() for const in self.constraints],
            "objective": self.objective.to_json()
        }
        return json_obj

    @classmethod
    def from_json(cls, json_obj):
        """
        Constructs a Model from the provided json-object.

        Example
        --------
        >>> import json
        >>> with open("path_to_file.json") as infile:
        >>>     model = Model.from_json(json.load(infile))
        """
        model = cls()
        model._init_from_json(json_obj)
        return model

    def _init_from_json(self, json_obj):
        self.name = json_obj["name"]
        interface = self.interface
        variables = [interface.Variable.from_json(var_json) for var_json in json_obj["variables"]]
        var_dict = {var.name: var for var in variables}
        constraints = [interface.Constraint.from_json(constraint_json, var_dict) for constraint_json in json_obj["constraints"]]
        objective = interface.Objective.from_json(json_obj["objective"], var_dict)
        self.add(variables)
        self.add(constraints)
        self.objective = objective
        self.update()

    def __getstate__(self):
        return self.to_json()

    def __setstate__(self, state):
        self.__init__()
        self._init_from_json(state)


if __name__ == '__main__':
    # Example workflow

    x1 = Variable('x1', lb=0)
    x2 = Variable('x2', lb=0)
    x3 = Variable('x3', lb=0)
    c1 = Constraint(x1 + x2 + x3, ub=100)
    c2 = Constraint(10 * x1 + 4 * x2 + 5 * x3, ub=600)
    c3 = Constraint(2 * x1 + 2 * x2 + 6 * x3, ub=300)
    obj = Objective(10 * x1 + 6 * x2 + 4 * x3, direction='max')
    model = Model(name='Simple model')
    model.objective = obj
    model.add([c1, c2, c3])

    try:
        sol = model.optimize()
    except NotImplementedError as e:
        print(e)

    print(model)
    print(model.variables)

    # model.remove(x1)

    model.interface = optlang.glpk_interface



# Copyright (c) 2016 Novo Nordisk Foundation Center for Biosustainability, DTU.
# See LICENSE for details.

import json
import unittest

import jsonschema
from optlang.glpk_interface import Variable, Constraint, Objective, Model

bound_schema = {
    "oneOf": [
        {"type": "number"},
        {"type": "null"}
    ]
}

variable_schema = {
    "$schema": "variable",
    "type": "object",
    "properties": {
        "name": {
            "type": "string",
        },
        "lb": bound_schema,
        "ub": bound_schema,
        "type": {
            "type": "string",
            "enum": ["continuous", "binary", "integer"]
        }
    },
    "required": ["name", "lb", "ub", "type"]
}

expression_definition = {
    "type": "object",
    "properties": {
        "name": {
            "type": "string"
        },
        "type": {
            "type": "string",
            "enum": ["Add", "Mul", "Pow", "Number", "Symbol"]
        },
        "value": {
            "type": "number"
        },
        "args": {
            "type": "array",
            "items": {"$ref": "#/definitions/expr"}
        }
    },
    "required": ["type"]
}

constraint_schema = {
    "type": "object",
    "properties": {
        "name": {
            "type": "string"
        },
        "lb": bound_schema,
        "ub": bound_schema,
        "indicator_variable": {
            "oneOf": [
                {"type": "string"},
                {"type": "null"}
            ]
        },
        "active_when": {
            "oneOf": [
                {"type": "number",
                 "enum": [0, 1]},
                {"type": "null"}
            ]
        },
        "expression": {"$ref": "#/definitions/expr"}
    },
    "definitions": {
        "expr": expression_definition
    }
}

objective_schema = {
    "type": "object",
    "properties": {
        "name": {
            "type": "string"
        },
        "direction": {
            "type": "string",
            "enum": ["min", "max"]
        },
        "expression": {"$ref": "#/definitions/expr"}
    },
    "definitions": {
        "expr": expression_definition
    }
}

model_schema = {
    "type": "object",
    "properties": {
        "name": {
            "oneOf": [
                {"type": "string"},
                {"type": "null"}
            ]
        },
        "objective": objective_schema,
        "variables": {
            "type": "array",
            "items": variable_schema
        },
        "constraints": {
            "type": "array",
            "items": constraint_schema
        }
    },
    "definitions": {
        "expr": expression_definition
    }
}


class JsonTest(unittest.TestCase):
    def setUp(self):
        self.var1 = var1 = Variable("var1", lb=0, ub=1, type="continuous")
        self.var2 = var2 = Variable("var2", lb=0, ub=1, type="continuous")
        self.const1 = const1 = Constraint(0.5 * var1, lb=0, ub=1, name="c1")
        self.const2 = const2 = Constraint(0.1 * var2 + 0.4 * var1, name="c2")
        self.model = model = Model()
        model.add([var1, var2])
        model.add([const1, const2])
        model.objective = Objective(var1 + var2)
        model.update()
        self.json_string = json.dumps(model.to_json())

    def test_model_json_validates(self):
        jsonschema.validate(self.model.to_json(), model_schema)

    def test_objective_json_validates(self):
        jsonschema.validate(self.model.objective.to_json(), objective_schema)

    def test_model_is_reconstructed_from_json(self):
        model = Model.from_json(json.loads(self.json_string))
        self.assertEqual(model.variables["var1"].lb, 0)
        self.assertEqual(model.variables["var1"].ub, 1)
        self.assertEqual(model.constraints["c1"].expression, 0.5 * model.variables["var1"])
        self.assertEqual(
            (model.objective.expression - sum(model.variables)).simplify(), 0
        )


if __name__ == "__main__":
    import nose

    nose.runmodule()



# Copyright (c) 2013 Novo Nordisk Foundation Center for Biosustainability, DTU.
# See LICENSE for details.

import unittest
import json
import os
import optlang.interface
import pickle
import copy
import sys

try:
    import mip
except ImportError as e:

    if str(e).find('mip') >= 0:
        class TestMissingDependency(unittest.TestCase):

            @unittest.skip('Missing dependency - ' + str(e))
            def test_fail(self):
                pass
    else:
        raise

else:

    from optlang import coinor_cbc_interface
    from optlang.tests import abstract_test_cases

    TESTMODELPATH = os.path.join(os.path.dirname(__file__), 'data/coli_core.json')

    def same_ex(ex1, ex2):
        """Compare to expressions for mathematical equality."""
        return ex1.simplify() == ex2.simplify()

    class VariableTestCase(abstract_test_cases.AbstractVariableTestCase):
        interface = coinor_cbc_interface

        def test_get_primal(self):
            self.assertEqual(self.var.primal, None)
            with open(TESTMODELPATH) as infile:
                model = self.interface.Model.from_json(json.load(infile))

            model.optimize()
            self.assertEqual(model.status, optlang.interface.OPTIMAL)
            for var in model.variables:
                self.assertTrue(var.lb <= round(var.primal, 6) <= var.ub, (var.lb, var.primal, var.ub))

        @unittest.skip("COIN-OR Cbc doesn't support variable name change")
        def test_changing_variable_names_is_reflected_in_the_solver(self):
            pass

        @unittest.skip("COIN-OR Cbc doesn't support variable name change")
        def test_change_name(self):
            pass

        def test_set_wrong_type_raises(self):
            self.assertRaises(ValueError, self.interface.Variable, name="test", type="mayo")
            self.assertRaises(Exception, setattr, self.var, 'type', 'ketchup')
            self.model.add(self.var)
            self.model.update()
            self.assertRaises(ValueError, setattr, self.var, "type", "mustard")

        def test_change_type(self):
            self.var.type = "continuous"
            self.assertEqual(self.var.lb, None)
            self.assertEqual(self.var.ub, None)
            self.var.type = "integer"
            self.assertEqual(self.var.lb, None)
            self.assertEqual(self.var.ub, None)
            self.var.type = "binary"
            self.assertEqual(self.var.lb, 0)
            self.assertEqual(self.var.ub, 1)
            self.var.type = "integer"
            self.assertEqual(self.var.lb, 0)
            self.assertEqual(self.var.ub, 1)
            self.var.type = "continuous"
            self.assertEqual(self.var.lb, 0)
            self.assertEqual(self.var.ub, 1)
            self.var.lb = -1.4
            self.var.ub = 1.6
            self.var.type = "integer"
            self.assertEqual(self.var.lb, -1)
            self.assertEqual(self.var.ub, 2)


    class ConstraintTestCase(abstract_test_cases.AbstractConstraintTestCase):
        interface = coinor_cbc_interface

        def test_get_primal(self):

            with open(TESTMODELPATH) as infile:
                self.model = self.interface.Model.from_json(json.load(infile))
            self.assertEqual(self.constraint.primal, None)
            self.model.optimize()
            for c in self.model.constraints:
                if c.lb is not None:
                    self.assertTrue(c.lb <= round(c.primal, 6))
                if c.ub is not None:
                    self.assertTrue(round(c.primal, 6) <= c.ub)

        @unittest.skip("COIN-OR Cbc doesn't support constraint name change")
        def test_change_constraint_name(self):
            pass

        @unittest.skip("TODO: Currently not supported")
        def test_indicator_constraint_support(self):
            pass


    class ObjectiveTestCase(abstract_test_cases.AbstractObjectiveTestCase):
        interface = coinor_cbc_interface

        def setUp(self):
            with open(TESTMODELPATH) as infile:
                self.model = self.interface.Model.from_json(json.load(infile))
            self.obj = self.model.objective

        def test_change_direction(self):
            from mip import MAXIMIZE, MINIMIZE
            self.obj.direction = "min"
            self.assertEqual(self.obj.direction, "min")
            self.assertEqual(self.model.problem.sense, MINIMIZE)

            self.obj.direction = "max"
            self.assertEqual(self.obj.direction, "max")
            self.assertEqual(self.model.problem.sense, MAXIMIZE)


    class ConfigurationTestCase(abstract_test_cases.AbstractConfigurationTestCase):
        interface = coinor_cbc_interface


    class ModelTestCase(abstract_test_cases.AbstractModelTestCase):
        interface = coinor_cbc_interface

        def setUp(self):
            with open(TESTMODELPATH) as infile:
                self.model = self.interface.Model.from_json(json.load(infile))

        def test_pickle_ability(self):
            self.model.optimize()
            value = self.model.objective.value
            pickle_string = pickle.dumps(self.model)
            from_pickle = pickle.loads(pickle_string)
            from_pickle.optimize()
            self.assertAlmostEqual(value, from_pickle.objective.value)
            self.assertEqual([(var.lb, var.ub, var.name, var.type) for var in from_pickle.variables.values()],
                             [(var.lb, var.ub, var.name, var.type) for var in self.model.variables.values()])
            self.assertEqual([(constr.lb, constr.ub, constr.name) for constr in from_pickle.constraints],
                             [(constr.lb, constr.ub, constr.name) for constr in self.model.constraints])

        def test_config_gets_copied_too(self):
            self.assertEqual(self.model.configuration.verbosity, 0)
            self.model.configuration.verbosity = 3
            model_copy = copy.copy(self.model)
            self.assertEqual(model_copy.configuration.verbosity, 3)

        def test_init_from_existing_problem(self):
            self.assertEqual(len(self.model.variables), len(self.model.problem.vars))
            # Divide by 2 because upper and lower constraints are represented seperately
            self.assertEqual(len(self.model.constraints), len(self.model.problem.constrs) / 2)
            self.assertEqual(self.model.variables.keys(),
                             [var.name[2:] for var in self.model.problem.vars])
            # Collect _lower and _upper constraints
            constrs= []
            for con in self.model.constraints:
                constrs.append(con.constraint_name(True))
                constrs.append(con.constraint_name(False))

            self.assertEqual(constrs, [constr.name for constr in self.model.problem.constrs])

        def test_add_non_cplex_conform_variable(self):
            var = self.interface.Variable('12x!!@#5_3', lb=-666, ub=666)
            self.model.add(var)
            self.assertTrue(var in self.model.variables.values())
            self.assertEqual(self.model.variables['12x!!@#5_3'].lb, -666)
            self.assertEqual(self.model.variables['12x!!@#5_3'].ub, 666)
            repickled = pickle.loads(pickle.dumps(self.model))
            var_from_pickle = repickled.variables['12x!!@#5_3']
            self.assertTrue('v_' + var_from_pickle.name in [var.name for var in self.model.problem.vars])

        @unittest.skip("COIN-OR Cbc doesn't support constraint name change")
        def test_change_constraint_name(self):
            pass

        def test_clone_model_with_lp(self):
            self.assertEqual(self.model.configuration.verbosity, 0)
            self.model.configuration.verbosity = 3
            self.model.optimize()
            opt = self.model.objective.value
            cloned_model = self.interface.Model.clone(self.model, use_lp=True)
            self.assertEqual(cloned_model.configuration.verbosity, 3)
            self.assertEqual(len(cloned_model.variables), len(self.model.variables))
            for var in self.model.variables:
                self.assertTrue(var.name in cloned_model.variables)
                var_clone = cloned_model.variables[var.name]
                self.assertEqual(var_clone.lb, var.lb)
                self.assertEqual(var_clone.ub, var.ub)
            self.assertEqual(len(cloned_model.constraints), len(self.model.constraints))
            for con in self.model.constraints:
                self.assertTrue(con.name in cloned_model.constraints)
                con_clone = cloned_model.constraints[con.name]
                self.assertEqual(con_clone.lb, con.lb)
                self.assertEqual(con_clone.ub, con.ub)
            cloned_model.optimize()
            self.assertAlmostEqual(cloned_model.objective.value, opt)

        def test_clone_small_model_with_lp(self):
            x1 = self.interface.Variable('x1', lb=0)
            x2 = self.interface.Variable('x2', lb=0)
            x3 = self.interface.Variable('x3', lb=0)

            # A constraint is constructed from an expression of variables and a lower and/or upper bound (lb and ub).
            c1 = self.interface.Constraint(x1 + x2 + x3, ub=100, name='c1')
            c2 = self.interface.Constraint(10 * x1 + 4 * x2 + 5 * x3, ub=600, name='c2')
            c3 = self.interface.Constraint(2 * x1 + 2 * x2 + 6 * x3, ub=300, name='c3')

            # An objective can be formulated
            obj = self.interface.Objective(10 * x1 + 6 * x2 + 4 * x3, direction='max')

            # Variables, constraints and objective are combined in a Model object, which can subsequently be optimized.
            model = self.interface.Model(name='Simple model')
            model.objective = obj
            model.add([c1, c2, c3])
            model.update()

            self.assertEqual(model.configuration.verbosity, 0)
            model.configuration.verbosity = 3
            model.optimize()
            opt = model.objective.value
            cloned_model = self.interface.Model.clone(model, use_lp=True)
            self.assertEqual(cloned_model.configuration.verbosity, 3)
            self.assertEqual(len(cloned_model.variables), len(model.variables))
            for var in model.variables:
                self.assertTrue(var.name in cloned_model.variables)
                var_clone = cloned_model.variables[var.name]
                self.assertEqual(var_clone.lb, var.lb)
                self.assertEqual(var_clone.ub, var.ub)
            self.assertEqual(len(cloned_model.constraints), len(model.constraints))
            for con in model.constraints:
                self.assertTrue(con.name in cloned_model.constraints)
                con_clone = cloned_model.constraints[con.name]
                self.assertEqual(con_clone.lb, con.lb)
                self.assertEqual(con_clone.ub, con.ub)
            cloned_model.optimize()
            self.assertAlmostEqual(cloned_model.objective.value, opt)

        def test_change_of_constraint_is_reflected_in_low_level_solver(self):
            x = self.interface.Variable('x', lb=0, ub=1, type='continuous')
            y = self.interface.Variable('y', lb=-181133.3, ub=12000., type='continuous')
            z = self.interface.Variable('z', lb=0., ub=10., type='continuous')
            constr1 = self.interface.Constraint(0.3 * x + 0.4 * y + 66. * z, lb=-100, ub=0., name='test')
            self.model.add(constr1)
            self.model.update()
            self.assertEqual(self.model.problem.constr_by_name('c_test_lower').rhs, 100)
            self.assertEqual(self.model.problem.constr_by_name('c_test_upper').rhs, 0)
            constr1.lb = -9
            constr1.ub = 10
            self.assertEqual(self.model.problem.constr_by_name('c_test_lower').rhs, 9)
            self.assertEqual(self.model.problem.constr_by_name('c_test_upper').rhs, 10)
            self.model.optimize()
            constr1.lb = -90
            constr1.ub = 100
            self.assertEqual(self.model.problem.constr_by_name('c_test_lower').rhs, 90)
            self.assertEqual(self.model.problem.constr_by_name('c_test_upper').rhs, 100)

        def test_constraint_set_problem_to_None_caches_the_latest_expression_from_solver_instance(self):
            x = self.interface.Variable('x', lb=-83.3, ub=1324422.)
            y = self.interface.Variable('y', lb=-181133.3, ub=12000.)
            constraint = self.interface.Constraint(0.3 * x + 0.4 * y, lb=-100, name='test')
            self.model.add(constraint)
            z = self.interface.Variable('z', lb=2, ub=5, type='integer')
            constraint += 77. * z
            self.model.remove(constraint)
            self.assertEqual(
                (constraint.expression - (0.4 * y + 0.3 * x + 77.0 * z)).expand() - 0, 0
            )
            self.assertEqual(constraint.lb, -100)
            self.assertEqual(constraint.ub, None)

        def test_change_of_objective_is_reflected_in_low_level_solver(self):
            x = self.interface.Variable('x', lb=-83.3, ub=1324422.)
            y = self.interface.Variable('y', lb=-181133.3, ub=12000.)
            objective = self.interface.Objective(0.3 * x + 0.4 * y, name='test', direction='max')
            self.model.objective = objective
            self.model.update()
            grb_x = self.model.problem.var_by_name('v_' + x.name)
            grb_y = self.model.problem.var_by_name('v_' + y.name)
            expected = {grb_x: 0.3, grb_y: 0.4}

            self.assertEqual(self.model.problem.objective.expr, expected)

            z = self.interface.Variable('z', lb=4, ub=4, type='integer')
            self.model.objective += 77. * z
            self.model.update()
            grb_z = self.model.problem.var_by_name('v_' + z.name)
            expected[grb_z] = 77.

            self.assertEqual(self.model.problem.objective.expr, expected)

        def test_change_variable_bounds(self):
            import random
            inner_prob = self.model.problem
            inner_problem_bounds = [(var.lb, var.ub) for var in inner_prob.vars]
            bounds = [(var.lb, var.ub) for var in self.model.variables.values()]
            self.assertEqual(bounds, inner_problem_bounds)
            for var in self.model.variables.values():
                var.ub = random.uniform(var.lb, 1000)
                var.lb = random.uniform(-1000, var.ub)
            self.model.update()
            inner_problem_bounds_new = [(var.lb, var.ub) for var in inner_prob.vars]
            bounds_new = [(var.lb, var.ub) for var in self.model.variables.values()]
            self.assertNotEqual(bounds, bounds_new)
            self.assertNotEqual(inner_problem_bounds, inner_problem_bounds_new)
            self.assertEqual(bounds_new, inner_problem_bounds_new)

        def test_change_constraint_bounds(self):
            constraint = self.model.constraints[0]
            value = 42
            constraint.ub = value
            self.assertEqual(constraint.ub, value)
            constraint.lb = value
            self.assertEqual(constraint.lb, value)
            name = constraint.name
            self.assertEqual(self.model.problem.constr_by_name('c_' + name + '_upper').rhs, value)
            self.assertEqual(self.model.problem.constr_by_name('c_' + name + '_lower').rhs, -1*value)

        def test_initial_objective(self):
            self.assertIn('BIOMASS_Ecoli_core_w_GAM', self.model.objective.expression.__str__(), )
            self.assertEqual(
                (self.model.objective.expression - (
                    1.0 * self.model.variables.BIOMASS_Ecoli_core_w_GAM -
                    1.0 * self.model.variables.BIOMASS_Ecoli_core_w_GAM_reverse_712e5)).expand() - 0, 0
            )

        def test_change_objective(self):
            v1, v2 = self.model.variables.values()[0:2]

            self.model.objective = self.interface.Objective(1. * v1 + 1. * v2)
            self.assertIn(v1.name, str(self.model.objective))
            self.assertIn(v2.name, str(self.model.objective))
            self.assertTrue(same_ex(self.model.objective._expression, 1.*v1 + 1.*v2))

            self.model.objective = self.interface.Objective(v1 + v2)
            self.assertIn(v1.name, str(self.model.objective))
            self.assertIn(v2.name, str(self.model.objective))
            self.assertTrue(same_ex(self.model.objective._expression, 1.*v1 + 1.*v2))

        def test_iadd_objective(self):
            v2, v3 = self.model.variables.values()[1:3]
            obj_coeff = sorted(self.model.problem.objective.expr.values())
            self.assertEqual(obj_coeff, [-1.0, 1.0])
            self.model.objective += 2. * v2 - 3. * v3
            obj_coeff = sorted(self.model.problem.objective.expr.values())
            self.assertEqual(obj_coeff, [-3.0, -1.0, 1.0, 2.0])

        def test_imul_objective(self):
            self.model.objective *= 2.
            obj_coeff = sorted(self.model.problem.objective.expr.values())
            self.assertEqual(obj_coeff, [-2.0, 2.0])

            v2, v3 = self.model.variables.values()[1:3]

            self.model.objective += 4. * v2 - 3. * v3
            self.model.objective *= 3.
            obj_coeff = sorted(self.model.problem.objective.expr.values())
            self.assertEqual(obj_coeff, [-9.0, -6.0, 6.0, 12.0])

            self.model.objective *= -1
            obj_coeff = sorted(self.model.problem.objective.expr.values())
            self.assertEqual(obj_coeff, [-12.0, -6.0, 6.0, 9.0])

        def test_set_copied_objective(self):
            mip_expr= self.model.problem.objective.expr
            obj_copy = copy.copy(self.model.objective)
            self.model.objective = obj_copy
            self.assertEqual(self.model.objective.direction, "max")
            self.assertEqual(mip_expr, self.model.problem.objective.expr)

        @unittest.skip("CBC-MIP timeout is flaky around 0")
        def test_timeout(self):
            pass

        def test_set_linear_coefficients_objective(self):
            self.model.objective.set_linear_coefficients({self.model.variables.BIOMASS_Ecoli_core_w_GAM: 666.})
            var = self.model.problem.var_by_name('v_' + self.model.variables.BIOMASS_Ecoli_core_w_GAM.name)
            self.assertEqual(self.model.problem.objective.expr[var], 666.)

        def test_set_linear_coefficients_constraint(self):
            constraint = self.model.constraints[0]
            coeff_dict = constraint.expression.as_coefficients_dict()
            self.assertEqual(coeff_dict[self.model.variables.GAPD_reverse_459c1], -1.0)
            constraint.set_linear_coefficients({self.model.variables.GAPD_reverse_459c1: 666.})
            coeff_dict = constraint.expression.as_coefficients_dict()
            self.assertEqual(coeff_dict[self.model.variables.GAPD_reverse_459c1], 666.)

        def test_coinor_cbc_coefficient_dict(self):
            x = self.interface.Variable("x")
            c = self.interface.Constraint(2 ** x, lb=0, sloppy=True)
            obj = self.interface.Objective(2 ** x, sloppy=True)
            model = self.interface.Model()
            self.assertRaises(Exception, setattr, model, "objective", obj)
            self.assertRaises(Exception, model._add_constraint, c)

            c = self.interface.Constraint(0, lb=0)
            obj = self.interface.Objective(0)
            model.add(c)
            model.objective = obj
            self.assertEqual(model.optimize(), optlang.interface.OPTIMAL)

        @unittest.skip('TODO: fix. Not working correctly')
        def test_integer_variable_dual(self):
            from functools import partial
            model = self.interface.Model()
            x = self.interface.Variable("x", lb=0)
            y = self.interface.Variable("y", lb=0)
            c = self.interface.Constraint(x + y, ub=1)
            model.add(c)
            model.objective = self.interface.Objective(x)

            model.optimize()
            self.assertEqual(y.dual, -1)

            x.type = "integer"
            model.optimize()
            # TODO: investigate. abstract test case has y. What should the
            #       behavior be?
            self.assertRaises(ValueError, partial(getattr, x, "dual"))

            x.type = "continuous"
            model.optimize()
            self.assertEqual(y.dual, -1)
            self.assertEqual(x.dual, 0)

        @unittest.skip('TODO: fix. Not working correctly')
        def test_integer_constraint_dual(self):
            pass

        @unittest.skip('TODO: fix. Not working correctly')
        def test_integer_batch_duals(self):
            pass

        def test_relax_with_knapsack(self):

            p = [10, 13, 18, 31, 7, 15]
            w = [11, 15, 20, 35, 10, 33]
            c, I = 47, range(len(w))

            x = [self.interface.Variable(type='binary', name='x{}'.format(i)) for i in I]

            obj = self.interface.Objective(sum(p[i] * x[i] for i in I), direction='max')

            c1 = self.interface.Constraint(sum(w[i] * x[i] for i in I), ub=c)

            model = self.interface.Model(name='knapsack')
            model.objective = obj
            model.add([c1])

            model.configuration.relax = True

            status = model.optimize()
            self.assertTrue(model.problem.relax)
            self.assertEqual(model.status, 'optimal')
            self.assertTrue(model.objective.value >= 41.0)


        def test_max_nodes_max_solutions_with_knapsack(self):

            p = [10, 13, 18, 31, 7, 15]
            w = [11, 15, 20, 35, 10, 33]
            c, I = 47, range(len(w))

            x = [self.interface.Variable(type='binary', name='x{}'.format(i)) for i in I]

            obj = self.interface.Objective(sum(p[i] * x[i] for i in I), direction='max')

            c1 = self.interface.Constraint(sum(w[i] * x[i] for i in I), ub=c)

            model = self.interface.Model(name='knapsack')
            model.objective = obj
            model.add([c1])

            model.configuration.max_nodes = 0
            model.configuration.max_solutions = 0
            status = model.optimize()
            self.assertEqual(model.problem.max_nodes, 0)
            self.assertEqual(model.problem.max_solutions, 0)
            self.assertEqual(model.status, 'feasible')

            model.configuration.max_solutions = 10
            status = model.optimize()
            self.assertEqual(model.problem.max_solutions, 10)
            self.assertEqual(model.status, 'optimal')

        def test_threads_cuts_emphasis_with_knapsack(self):

            p = [10, 13, 18, 31, 7, 15]
            w = [11, 15, 20, 35, 10, 33]
            c, I = 47, range(len(w))

            x = [self.interface.Variable(type='binary', name='x{}'.format(i)) for i in I]

            obj = self.interface.Objective(sum(p[i] * x[i] for i in I), direction='max')

            c1 = self.interface.Constraint(sum(w[i] * x[i] for i in I), ub=c)

            model = self.interface.Model(name='knapsack')
            model.objective = obj
            model.add([c1])

            model.configuration.threads = -1
            model.configuration.cuts = 1
            model.configuration.emphasis = 2

            status = model.optimize()
            self.assertEqual(model.problem.threads, -1)
            self.assertEqual(model.problem.cuts, 1)
            self.assertEqual(model.problem.emphasis, 2)
            self.assertEqual(model.status, 'optimal')



    class MIPExamples(unittest.TestCase):
        interface = coinor_cbc_interface

        def test_constant_objective(self):
            x1 = self.interface.Variable('x1', lb=0, ub=5)
            c1 = self.interface.Constraint(x1, lb=-10, ub=10, name='c1')
            obj = self.interface.Objective(1)
            model = self.interface.Model()
            model.objective = obj
            model.add(c1)
            model.optimize()

            self.assertEqual(model.status, 'optimal')
            self.assertEqual(model.objective.value, 1.0)

        def test_knapsack(self):

            p = [10, 13, 18, 31, 7, 15]
            w = [11, 15, 20, 35, 10, 33]
            c, I = 47, range(len(w))

            x = [self.interface.Variable(type='binary', name='x{}'.format(i)) for i in I]

            obj = self.interface.Objective(sum(p[i] * x[i] for i in I), direction='max')

            c1 = self.interface.Constraint(sum(w[i] * x[i] for i in I), ub=c)

            model = self.interface.Model(name='knapsack')
            model.objective = obj
            model.add([c1])

            status = model.optimize()

            self.assertEqual(model.status, 'optimal')
            self.assertEqual(model.objective.value, 41.0)

            primal_values = [val for val in model.primal_values.values()]
            self.assertEqual(primal_values, [1, 0, 0, 1, 0, 0])

            selected = [i for i in I if x[i].primal >= 0.99]
            self.assertEqual(selected, [0, 3])

        def test_travelling_salesman(self):
            from itertools import product

            # names of places to visit
            places = ['Antwerp', 'Bruges', 'C-Mine', 'Dinant', 'Ghent',
                      'Grand-Place de Bruxelles', 'Hasselt', 'Leuven',
                      'Mechelen', 'Mons', 'Montagne de Bueren', 'Namur',
                      'Remouchamps', 'Waterloo']

            # distances in an upper triangular matrix
            dists = [[83, 81, 113, 52, 42, 73, 44, 23, 91, 105, 90, 124, 57],
                     [161, 160, 39, 89, 151, 110, 90, 99, 177, 143, 193, 100],
                     [90, 125, 82, 13, 57, 71, 123, 38, 72, 59, 82],
                     [123, 77, 81, 71, 91, 72, 64, 24, 62, 63],
                     [51, 114, 72, 54, 69, 139, 105, 155, 62],
                     [70, 25, 22, 52, 90, 56, 105, 16],
                     [45, 61, 111, 36, 61, 57, 70],
                     [23, 71, 67, 48, 85, 29],
                     [74, 89, 69, 107, 36],
                     [117, 65, 125, 43],
                     [54, 22, 84],
                     [60, 44],
                     [97],
                     []]

            # number of nodes and list of vertices
            n, V = len(dists), set(range(len(dists)))

            # distances matrix
            c = [[0 if i == j
                  else dists[i][j-i-1] if j > i
                  else dists[j][i-j-1]
                  for j in V] for i in V]

            # binary variables indicating if arc (i,j) is used on the route or not
            x = [[self.interface.Variable(type='binary', name='x_i={}_j={}_arc'.format(i, j)) for j in V] for i in V]

            # continuous variable to prevent subtours: each city will have a
            # different sequential id in the planned route except the first one
            y = [self.interface.Variable(name='x{}'.format(i)) for i in V]

            # objective function: minimize the distance
            obj = self.interface.Objective(sum(c[i][j]*x[i][j] for i in V for j in V), direction='min')

            # constraint : leave each city only once
            cons = []
            for i in V:
                cons.append(self.interface.Constraint(sum(x[i][j] for j in V - {i}), lb=1, ub=1))

            # constraint : enter each city only once
            for i in V:
                cons.append(self.interface.Constraint(sum(x[j][i] for j in V - {i}), lb=1, ub=1))

            # subtour elimination
            for (i, j) in product(V - {0}, V - {0}):
                if i != j:
                    cons.append(self.interface.Constraint(y[i] - (n+1)*x[i][j] - y[j], lb=-1*n))

            model = self.interface.Model(name='travelling_salesman')
            model.objective = obj
            model.add(cons)
            model.optimize()

            self.assertEqual(model.status, 'optimal')
            self.assertEqual(model.objective.value, 547.0)



# Copyright 2013 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import optlang

# This function is very complex. Should maybe be refactored
def convert_linear_problem_to_dual(model, sloppy=False, infinity=None, maintain_standard_form=True, prefix="dual_", dual_model=None):  # NOQA
    """
    A mathematical optimization problem can be viewed as a primal and a dual problem. If the primal problem is
    a minimization problem the dual is a maximization problem, and the optimal value of the dual is a lower bound of
    the optimal value of the primal.
    For linear problems, strong duality holds, which means that the optimal values of the primal and dual are equal
    (duality gap = 0).

    This functions takes an optlang Model representing a primal linear problem and returns a new Model representing
    the dual optimization problem. The provided model must have a linear objective, linear constraints and only
    continuous variables. Furthermore, the problem must be in standard form, i.e. all variables should be non-negative.
    Both minimization and maximization problems are allowed. The objective direction of the dual will always be
    opposite of the primal.

    Attributes:
    ----------
    model: optlang.interface.Model
        The primal problem to be dualized
    sloppy: Boolean (default False)
        If True, linearity, variable types and standard form will not be checked. Only use if you know the primal is
        valid
    infinity: Numeric or None
        If not None this value will be used as bounds instead of unbounded variables.
    maintain_standard_form: Boolean (default True)
        If False the returned dual problem will not be in standard form, but will have fewer variables and/or constraints
    prefix: str
        The string that will be prepended to all variable and constraint names in the returned dual problem.
    dual_model: optlang.interface.Model or None (default)
        If not None, the dual variables and constraints will be added to this model. Note the objective will also be
        set to the dual objective. If None a new model will be created.

    Returns:
    ----------
    dual_problem: optlang.interface.Model (same solver as the primal)
    """
    if dual_model is None:
        dual_model = model.interface.Model()

    maximization = model.objective.direction == "max"

    if infinity is not None:
        neg_infinity = -infinity
    else:
        neg_infinity = None

    if maximization:
        sign = 1
    else:
        sign = -1

    coefficients = {}
    dual_objective = {}

    # Add dual variables from primal constraints:
    for constraint in model.constraints:
        if constraint.expression == 0:
            continue  # Skip empty constraint
        if not (sloppy or constraint.is_Linear):
            raise ValueError("Non-linear problems are not supported: " + str(constraint))
        if constraint.lb is None and constraint.ub is None:
            continue  # Skip free constraint
        if not maintain_standard_form and constraint.lb == constraint.ub:
            const_var = model.interface.Variable(prefix + constraint.name + "_constraint", lb=neg_infinity, ub=infinity)
            dual_model.add(const_var)
            if constraint.lb != 0:
                dual_objective[const_var] = sign * constraint.lb
            for variable, coef in constraint.expression.as_coefficients_dict().items():
                if variable == 1:  # pragma: no cover  # For symengine
                    continue
                coefficients.setdefault(variable.name, {})[const_var] = sign * coef
        else:
            if constraint.lb is not None:
                lb_var = model.interface.Variable(prefix + constraint.name + "_constraint_lb", lb=0, ub=infinity)
                dual_model.add(lb_var)
                if constraint.lb != 0:
                    dual_objective[lb_var] = -sign * constraint.lb
            if constraint.ub is not None:
                ub_var = model.interface.Variable(prefix + constraint.name + "_constraint_ub", lb=0, ub=infinity)
                dual_model.add(ub_var)
                if constraint.ub != 0:
                    dual_objective[ub_var] = sign * constraint.ub

            assert constraint.expression.is_Add or constraint.expression.is_Mul, \
                "Invalid expression type: " + str(type(constraint.expression))
            if constraint.expression.is_Add:
                coefficients_dict = constraint.expression.as_coefficients_dict()
            else:  # constraint.expression.is_Mul:
                coefficients_dict = {constraint.expression.args[1]: constraint.expression.args[0]}

            for variable, coef in coefficients_dict.items():
                if variable == 1:  # pragma: no cover  # For symengine
                    continue
                if constraint.lb is not None:
                    coefficients.setdefault(variable.name, {})[lb_var] = -sign * coef
                if constraint.ub is not None:
                    coefficients.setdefault(variable.name, {})[ub_var] = sign * coef

    # Add dual variables from primal bounds
    for variable in model.variables:
        if not (sloppy or variable.type == "continuous"):
            raise ValueError("Integer variables are not supported: " + str(variable))
        if not sloppy and (variable.lb is None or variable.lb < 0):
            raise ValueError("Problem is not in standard form (" + variable.name + " can be negative)")
        if variable.lb > 0:
            bound_var = model.interface.Variable(prefix + variable.name + "_lb", lb=0, ub=infinity)
            dual_model.add(bound_var)
            coefficients.setdefault(variable.name, {})[bound_var] = -sign * 1
            dual_objective[bound_var] = -sign * variable.lb
        if variable.ub is not None:
            bound_var = model.interface.Variable(prefix + variable.name + "_ub", lb=0, ub=infinity)
            dual_model.add(bound_var)
            coefficients.setdefault(variable.name, {})[bound_var] = sign * 1
            if variable.ub != 0:
                dual_objective[bound_var] = sign * variable.ub

    # Add dual constraints from primal objective
    primal_objective_dict = model.objective.expression.as_coefficients_dict()
    for variable in model.variables:
        expr = optlang.symbolics.add([(coef * dual_var) for dual_var, coef in coefficients[variable.name].items()])
        obj_coef = primal_objective_dict[variable]
        if maximization:
            const = model.interface.Constraint(expr, lb=obj_coef, name=prefix + variable.name)
        else:
            const = model.interface.Constraint(expr, ub=obj_coef, name=prefix + variable.name)
        dual_model.add(const)

    # Make dual objective
    expr = optlang.symbolics.add([(coef * dual_var) for dual_var, coef in dual_objective.items() if coef != 0])
    if maximization:
        objective = model.interface.Objective(expr, direction="min")
    else:
        objective = model.interface.Objective(expr, direction="max")
    dual_model.objective = objective

    return dual_model



# Copyright 2015 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The scipy interface uses the 'linprog' function of the scipy package.
It has no integer or QP capabilities. Furthermore dual values (reduced costs and shadow prices) are
not supported by this interface.

This interface works well with small to medium scale models but for better performance, other
solvers should be used for large models.
"""


from __future__ import absolute_import, print_function

from collections import OrderedDict
from itertools import islice

import numpy as np
import six
from optlang import interface
from optlang.expression_parsing import parse_optimization_expression
from scipy.optimize import linprog
from optlang.util import inheritdocstring
from optlang import symbolics

SCIPY_STATUS = {
    0: interface.OPTIMAL,
    1: interface.ABORTED,
    2: interface.INFEASIBLE,
    3: interface.UNBOUNDED
}


class Problem(object):
    """
    Scipy linprog problem. This object implements an object-oriented interface to the linprog solver function.
    This class is wrapped by the Model object, but can also be used as a stand-alone low-level interface.
    """
    def __init__(self):
        self._objective = OrderedDict()
        self._direction = "min"

        self._rows_to_be_added = None
        self._cols_to_be_added = None

        self._A = np.zeros(shape=[0, 0])
        self.upper_bounds = np.zeros([0])
        self.bounds = OrderedDict()

        self._variables = OrderedDict()

        self._constraints = OrderedDict()

        self._status = None
        self._var_primals = None
        self._slacks = None
        self._f = None

    def _get_var_index(self, name):
        return self._variables[name]

    def _get_constraint_index(self, name):
        return self._constraints[name]

    def _add_row_to_A(self, row):
        # Flush any cols_to_be_added
        # if self._cols_to_be_added is not None:
        #    self._flush_cols_to_add()
        # Add the row to queue
        # if self._rows_to_be_added is None:
        #    self._rows_to_be_added = []
        # self._rows_to_be_added.append(row)
        self._A = np.vstack((self._A, row))

    # def _flush_rows_to_add(self):
    #     self._A = np.vstack([self._A] + self._rows_to_be_added)
    #     self._rows_to_be_added = None

    def _add_col_to_A(self, col):
        # Flush any rows_to_be_added
        # if self._rows_to_be_added is not None:
        #    self._flush_rows_to_add()
        # Add the col to queue
        # if self._cols_to_be_added is None:
        #    self._cols_to_be_added = []
        # self._cols_to_be_added.append(col)
        self._A = np.hstack((self._A, col))

    # def _flush_cols_to_add(self):
    #     self._A = np.hstack([self._A] + self._cols_to_be_added)
    #     self._cols_to_be_added = None

    def set_variable_bounds(self, name, lower, upper):
        """Set the bounds of a variable"""
        self.bounds[name] = (lower, upper)
        self._reset_solution()

    def add_variable(self, name):
        """Add a variable to the problem"""
        if name in self._variables:
            raise ValueError(
                "A variable named " + name + " already exists."
            )
        self._variables[name] = len(self._variables)
        self.bounds[name] = (0, None)

        new_col = np.zeros(shape=[len(self._constraints), 1])
        self._add_col_to_A(new_col)
        self._reset_solution()

    def add_constraint(self, name, coefficients={}, ub=0):
        """
        Add a constraint to the problem. The constrain is formulated as a dictionary of variable names to
        linear coefficients.
        The constraint can only have an upper bound. To make a constraint with a lower bound, multiply
        all coefficients by -1.
        """
        if name in self._constraints:
            raise ValueError(
                "A constraint named " + name + " already exists."
            )
        self._constraints[name] = len(self._constraints)
        self.upper_bounds = np.append(self.upper_bounds, ub)

        new_row = np.array([[coefficients.get(name, 0) for name in self._variables]])
        self._add_row_to_A(new_row)
        self._reset_solution()

    def change_var_name(self, name, value):
        if name != value and value in self._variables:
            raise ValueError("A variable with that name already exists")
        self._variables = OrderedDict(((value if k == name else k), v) for k, v in self._variables.items())
        self._objective = OrderedDict(((value if k == name else k), v) for k, v in self._objective.items())
        self.bounds = OrderedDict(((value if k == name else k), v) for k, v in self.bounds.items())

    def change_constraint_name(self, name, value):
        if name != value and value in self._constraints:
            raise ValueError("A constraint with that name already exists")
        self._constraints = OrderedDict(((value if k == name else k), v) for k, v in self._constraints.items())

    def remove_variable(self, name):
        """Remove a variable from the problem."""
        index = self._get_var_index(name)
        # Remove from matrix
        self._A = np.delete(self.A, index, 1)
        # Remove from bounds
        del self.bounds[name]
        # Remove from var list
        del self._variables[name]
        self._update_variable_indices()
        self._reset_solution()

    def remove_constraint(self, name):
        """Remove a constraint from the problem"""
        index = self._get_constraint_index(name)
        # Remove from matrix
        self._A = np.delete(self.A, index, 0)
        # Remove from upper_bounds
        self.upper_bounds = np.delete(self.upper_bounds, index)
        # Remove from constraint list
        del self._constraints[name]
        self._update_constraint_indices()
        self._reset_solution()

    def set_constraint_bound(self, name, value):
        """Set the upper bound of a constraint."""
        index = self._get_constraint_index(name)
        self.upper_bounds[index] = value
        self._reset_solution()

    def get_var_primal(self, name):
        """Get the primal value of a variable. Returns None if the problem has not bee optimized."""
        if self._var_primals is None:
            return None
        else:
            index = self._get_var_index(name)
            return self._var_primals[index]

    def get_var_dual(self, name):
        raise NotImplementedError(
            "Duals are currently not implemented for the Scipy solver. Please use another interface instead.")

    @property
    def A(self):
        """The linear coefficient matrix."""
        assert self._rows_to_be_added is None or self._cols_to_be_added is None
        # if self._rows_to_be_added is not None:
        #     self._flush_rows_to_add()
        # if self._cols_to_be_added is not None:
        #     self._flush_cols_to_add()
        return self._A

    @property
    def objective(self):
        return self._objective

    @objective.setter
    def objective(self, value):
        self._objective = value
        self._reset_solution()

    @property
    def offset(self):
        return getattr(self, '_offset', 0)

    @offset.setter
    def offset(self, value):
        self._offset = value

    @property
    def direction(self):
        return self._direction

    @direction.setter
    def direction(self, value):
        self._direction = value
        self._reset_solution()

    @property
    def status(self):
        if self._status is None:
            raise RuntimeError("Problem has not been optimized yet")
        return SCIPY_STATUS[self._status]

    def get_constraint_slack(self, name):
        """Get the value of the slack variable of a constraint."""
        if self._slacks is None:
            return None
        else:
            index = self._get_constraint_index(name)
            return self._slacks[index]

    def optimize(self, method="simplex", verbosity=False, tolerance=1e-9, **kwargs):
        """Run the linprog function on the problem. Returns None."""
        c = np.array([self.objective.get(name, 0) for name in self._variables])
        if self.direction == "max":
            c *= -1

        bounds = list(six.itervalues(self.bounds))
        solution = linprog(c, self.A, self.upper_bounds, bounds=bounds, method=method,
                           options={"maxiter": 10000, "disp": verbosity, "tol": tolerance}, **kwargs)
        self._solution = solution
        self._status = solution.status
        if SCIPY_STATUS[self._status] == interface.OPTIMAL:
            self._var_primals = solution.x
            self._slacks = solution.slack
        else:
            self._var_primals = None
            self._slacks = None

        self._f = solution.fun

    @property
    def objective_value(self):
        """Returns the optimal objective value"""
        if self._f is None:
            raise RuntimeError("Problem has not been optimized yet")
        if self.direction == "max":
            return -self._f + self.offset
        else:
            return self._f + self.offset

    def _update_variable_indices(self, start=0):
        i = start
        for name in islice(self._variables, start, len(self._constraints)):
            self._variables[name] = i
            i += 1

    def _update_constraint_indices(self, start=0):
        i = start
        for name in islice(self._constraints, start, len(self._constraints)):
            self._constraints[name] = i
            i += 1

    def _reset_solution(self):
        self._status = None
        self._var_primals = None
        self._slacks = None
        self._f = None


@six.add_metaclass(inheritdocstring)
class Variable(interface.Variable):
    def __init__(self, name, lb=None, ub=None, type="continuous", *args, **kwargs):
        if type != "continuous":
            raise ValueError("Scipy solver only works with continuous variables. Please use another interface")
        super(Variable, self).__init__(name, lb, ub, type, *args, **kwargs)

    @interface.Variable.lb.setter
    def lb(self, value):
        interface.Variable.lb.fset(self, value)
        if self.problem:
            self.problem.problem.set_variable_bounds(self.name, self.lb, self.ub)

    @interface.Variable.ub.setter
    def ub(self, value):
        interface.Variable.ub.fset(self, value)
        if self.problem:
            self.problem.problem.set_variable_bounds(self.name, self.lb, self.ub)

    def set_bounds(self, lb, ub):
        super(Variable, self).set_bounds(lb, ub)
        if self.problem:
            self.problem.problem.set_variable_bounds(self.name, lb, ub)

    @interface.Variable.type.setter
    def type(self, value):
        if value != "continuous":
            raise ValueError("Scipy solver only works with continuous variables. Please use another interface")
        super(Variable, Variable).type.fset(self, value)

    @property
    def primal(self):
        if getattr(self, "problem", None) is not None:
            primal = self.problem.problem.get_var_primal(self.name)
            return primal
        else:
            return None

    @property
    def dual(self):
        if getattr(self, "problem", None) is not None:
            dual = self.problem.problem.get_var_dual(self.name)
            return dual
        else:
            return None

    @interface.Variable.name.setter
    def name(self, value):
        if getattr(self, "problem", None) is not None:
            self.problem.problem.change_var_name(self.name, value)
        super(Variable, Variable).name.fset(self, value)


@six.add_metaclass(inheritdocstring)
class Constraint(interface.Constraint):
    _INDICATOR_CONSTRAINT_SUPPORT = False

    def __init__(self, expression, sloppy=False, *args, **kwargs):
        super(Constraint, self).__init__(expression, sloppy=sloppy, *args, **kwargs)
        if not sloppy:
            if not self.is_Linear:
                raise ValueError(
                    "Scipy only supports linear constraints. %s is not linear." % self)

    @property
    def upper_constraint_name(self):
        return self.name + "_upper"

    @property
    def lower_constraint_name(self):
        return self.name + "_lower"

    @property
    def primal(self):
        if getattr(self, "problem", None) is not None and self.problem.status == interface.OPTIMAL:
            if self.lb is not None:
                primal = self.lb + self.problem.problem.get_constraint_slack(self.lower_constraint_name)
                return primal
            elif self.ub is not None:
                primal = self.ub - self.problem.problem.get_constraint_slack(self.upper_constraint_name)
                return primal
            else:
                return None
        else:
            return None

    @property
    def dual(self):
        if getattr(self, "problem", None) is not None:
            raise NotImplementedError(
                "Duals have not been implemented for the Scipy solver. Please use another interface."
            )
        else:
            return None

    @interface.Constraint.lb.setter
    def lb(self, value):
        self._check_valid_lower_bound(value)
        if getattr(self, "problem", None) is not None:
            if self.lb is None and value is not None:
                negative_coefficient_dict = {name: -coef for name, coef in self.coefficient_dict().items()}
                self.problem.problem.add_constraint(
                    self.lower_constraint_name, negative_coefficient_dict, ub=-value
                )
            elif value is None and self.lb is not None:
                self.problem.problem.remove_constraint(self.lower_constraint_name)
            elif value is not None and self.lb is not None:
                self.problem.problem.set_constraint_bound(self.lower_constraint_name, -value)
        self._lb = value

    @interface.Constraint.ub.setter
    def ub(self, value):
        self._check_valid_upper_bound(value)
        if getattr(self, "problem", None) is not None:
            if self.ub is None and value is not None:
                self.problem.problem.add_constraint(self.upper_constraint_name, self.coefficient_dict(), ub=value)
            elif value is None and self.ub is not None:
                self.problem.problem.remove_constraint(self.upper_constraint_name)
            elif value is not None and self.ub is not None:
                self.problem.problem.set_constraint_bound(self.upper_constraint_name, value)
        self._ub = value

    def coefficient_dict(self, names=True):
        if self.expression.is_Add:
            coefficient_dict = {variable: coef for variable, coef in
                                self.expression.as_coefficients_dict().items() if variable.is_Symbol}
            coefficient_dict = {var: float(coef) for var, coef in coefficient_dict.items()}
        elif self.expression.is_Atom and self.expression.is_Symbol:
            coefficient_dict = {self.expression: 1}
        elif self.expression.is_Mul and len(self.expression.args) <= 2:
            args = self.expression.args
            coefficient_dict = {args[1]: float(args[0])}
        elif self.expression.is_Number:
            coefficient_dict = {}
        else:
            raise ValueError("Invalid expression: " + str(self.expression))
        if names is True:
            coefficient_dict = {var.name: coef for var, coef in coefficient_dict.items()}
        return coefficient_dict

    def set_linear_coefficients(self, coefficients):
        if self.problem is not None:
            self.problem.update()
            lb, ub = self.lb, self.ub
            self.lb, self.ub = None, None
            coefficients_dict = self.coefficient_dict(names=False)
            coefficients_dict.update(coefficients)
            self._expression = symbolics.add(*(v * k for k, v in coefficients_dict.items()))
            self.lb = lb
            self.ub = ub
        else:
            raise Exception("Can't change coefficients if constraint is not associated with a model.")

    def get_linear_coefficients(self, variables):
        if self.problem is not None:
            self.problem.update()
            coefs = self.coefficient_dict(names=False)
            return {v: coefs.get(v, 0) for v in variables}
        else:
            raise Exception("Can't get coefficients from solver if constraint is not in a model")


@six.add_metaclass(inheritdocstring)
class Objective(interface.Objective):
    def __init__(self, expression, sloppy=False, **kwargs):
        super(Objective, self).__init__(expression, sloppy=sloppy, **kwargs)
        if not (sloppy or self.is_Linear):
            raise ValueError(
                "Scipy only supports linear objectives. %s is not linear." % self)

    def _get_expression(self):
        if self.problem is not None:
            coefficients_dict = self.problem.problem.objective
            coefficients_dict = {
                self.problem._variables[name]: coef for name, coef in coefficients_dict.items() if name in self.problem._variables
            }
            self._expression = symbolics.add(*(v * k for k, v in coefficients_dict.items())) + self.problem.problem.offset
        return self._expression

    @property
    def value(self):
        if getattr(self, "problem", None) is not None:
            return self.problem.problem.objective_value
        else:
            return None

    @interface.Objective.direction.setter
    def direction(self, value):
        super(Objective, self.__class__).direction.fset(self, value)
        if getattr(self, "problem", None) is not None:
            self.problem.problem.direction = value

    def coefficient_dict(self):
        if self.expression.is_Add:
            coefficient_dict = {variable: coef for variable, coef in
                                self.expression.as_coefficients_dict().items() if variable.is_Symbol}
        elif self.expression.is_Atom and self.expression.is_Symbol:
            coefficient_dict = {self.expression: 1}
        elif self.expression.is_Mul and len(self.expression.args) <= 2 and self.expression.args[1].is_Symbol:
            args = self.expression.args
            coefficient_dict = {args[1]: float(args[0])}
        elif self.expression.is_Number:
            coefficient_dict = {}
        else:
            raise ValueError("Invalid expression: " + str(self.expression))
        coefficient_dict = {var.name: float(coef) for var, coef in coefficient_dict.items()}
        return coefficient_dict

    def set_linear_coefficients(self, coefficients):
        if self.problem is not None:
            self.problem.update()
            self.problem.problem.objective.update({var.name: coef for var, coef in coefficients.items()})
        else:
            raise Exception("Can't change coefficients if objective is not associated with a model.")

    def get_linear_coefficients(self, variables):
        if self.problem is not None:
            self.problem.update()
            return {v: float(self.problem.problem.objective.get(v.name, 0)) for v in variables}
        else:
            raise Exception("Can't get coefficients from solver if objective is not in a model")


@six.add_metaclass(inheritdocstring)
class Configuration(interface.MathematicalProgrammingConfiguration):
    def __init__(self, verbosity=0, tolerance=1e-9, *args, **kwargs):
        super(Configuration, self).__init__(*args, **kwargs)
        self._verbosity = verbosity
        self.tolerance = tolerance

    @property
    def verbosity(self):
        return self._verbosity

    @verbosity.setter
    def verbosity(self, value):
        self._verbosity = value

    @property
    def presolve(self):
        return False

    @presolve.setter
    def presolve(self, value):
        if value is not False:
            raise ValueError("The Scipy solver has no presolve capabilities")

    @property
    def timeout(self):
        return None

    @timeout.setter
    def timeout(self, value):
        if value is not None:
            raise ValueError("Scipy interface does not support timeout")

    @property
    def tolerance(self):
        return self._tolerance

    @tolerance.setter
    def tolerance(self, value):
        self._tolerance = value


@six.add_metaclass(inheritdocstring)
class Model(interface.Model):
    def _initialize_problem(self):
        self.problem = Problem()

    def _initialize_model_from_problem(self, problem):
        if isinstance(problem, Problem):
            self.problem = problem
        else:
            raise TypeError("Problem must be an instance of scipy_interface.Problem, not " + repr(type(problem)))

    # def _add_variable(self, variable):
    #     print("added", variable.name)
    #     super(Model, self)._add_variable(variable)
    #     self.problem.add_variable(variable.name)
    #     self.problem.set_variable_bounds(variable.name, variable.lb, variable.ub)

    def _add_variables(self, variables):
        super(Model, self)._add_variables(variables)
        for var in variables:
            self.problem.add_variable(var.name)
            self.problem.set_variable_bounds(var.name, var.lb, var.ub)

    def _remove_variables(self, variables):
        variable_names = [variable.name for variable in variables]
        super(Model, self)._remove_variables(variables)
        for name in variable_names:
            self.problem.remove_variable(name)

    def _add_constraints(self, constraints, sloppy=False):
        if sloppy is False:
            for c in constraints:
                if not c.is_Linear:
                    raise ValueError("Scipy solver only works with linear constraints. Please use another interface.")
        super(Model, self)._add_constraints(constraints, sloppy=sloppy)

        for constraint in constraints:
            coefficient_dict = constraint.coefficient_dict()

            if constraint.ub is not None:
                self.problem.add_constraint(constraint.upper_constraint_name, coefficient_dict)
                self.problem.set_constraint_bound(constraint.upper_constraint_name, constraint.ub)

            if constraint.lb is not None:
                negative_coefficient_dict = {name: -coef for name, coef in coefficient_dict.items()}
                self.problem.add_constraint(constraint.lower_constraint_name, negative_coefficient_dict)
                self.problem.set_constraint_bound(constraint.lower_constraint_name, -constraint.lb)

    def _remove_constraints(self, constraints):
        for constraint in constraints:
            if constraint.lb is not None:
                self.problem.remove_constraint(constraint.lower_constraint_name)
            if constraint.ub is not None:
                self.problem.remove_constraint(constraint.upper_constraint_name)

            super(Model, self)._remove_constraints(constraints)

    def _optimize(self):
        self.problem.optimize(verbosity=bool(self.configuration.verbosity), tolerance=self.configuration.tolerance)
        status = self.problem.status
        return status

    @interface.Model.objective.setter
    def objective(self, value):
        super(Model, Model).objective.fset(self, value)
        value.problem = None
        if value is None:
            self.problem.objective = {}
        else:
            offset, coefficients, _ = parse_optimization_expression(value)
            self.problem.objective = {var.name: coef for var, coef in coefficients.items()}
            self.problem.offset = offset
            self.problem.direction = value.direction
        value.problem = self


if __name__ == "__main__":
    x1 = Variable('x1', lb=0)
    x2 = Variable('x2', lb=0)
    x3 = Variable('x3', lb=0)
    c1 = Constraint(x1 + x2 + x3, lb=-100, ub=100, name='c1')
    c2 = Constraint(10 * x1 + 4 * x2 + 5 * x3, ub=600, name='c2')
    c3 = Constraint(2 * x1 + 2 * x2 + 6 * x3, ub=300, name='c3')
    obj = Objective(10 * x1 + 6 * x2 + 4 * x3, direction='max')
    model = Model(name='Simple model')
    model.objective = obj
    model.add([c1, c2, c3])
    status = model.optimize()
    print("status:", model.status)
    print("objective value:", model.objective.value)

    for var in model.variables:
        print(var.name, "=", var.primal)



# Copyright 2013 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Interface for the Python MIP (Mixed-Integer Linear Programming) Tools.

Wraps the MIP solver by subclassing and extending :class:`Model`,
:class:`Variable`, and :class:`Constraint` from :mod:`interface`.

MIP is an open source MILP Python wrapper around the open-source COIN-OR
Branch-&-Cut CBC solver. To use MIP you need to install the 'mip'
python package (with pip or from https://www.python-mip.com/)
and make sure that 'import mip' runs without error.
"""
import logging

import six

from optlang.util import inheritdocstring, TemporaryFilename
from optlang.expression_parsing import parse_optimization_expression
from optlang import interface
from optlang import symbolics

from math import isclose, ceil, floor

try:
    import mip
except ImportError:
    raise ImportError("The coinor_cbc_interface requires mip!")


log = logging.getLogger(__name__)

_MIP_STATUS_TO_STATUS = {
    mip.OptimizationStatus.CUTOFF: interface.CUTOFF,
    mip.OptimizationStatus.ERROR: interface.ABORTED,
    mip.OptimizationStatus.FEASIBLE: interface.FEASIBLE,
    mip.OptimizationStatus.INFEASIBLE: interface.INFEASIBLE,
    mip.OptimizationStatus.INT_INFEASIBLE: interface.SPECIAL,
    mip.OptimizationStatus.LOADED: interface.LOADED,
    mip.OptimizationStatus.NO_SOLUTION_FOUND: interface.NOFEASIBLE,
    mip.OptimizationStatus.OPTIMAL: interface.OPTIMAL,
    mip.OptimizationStatus.UNBOUNDED: interface.UNBOUNDED,
    mip.OptimizationStatus.OTHER: interface.SPECIAL
}

_MIP_VTYPE_TO_VTYPE = {
    mip.CONTINUOUS: 'continuous',
    mip.INTEGER: 'integer',
    mip.BINARY: 'binary'
}

_VTYPE_TO_MIP_VTYPE = dict(
    [(val, key) for key, val in six.iteritems(_MIP_VTYPE_TO_VTYPE)]
)

_DIR_TO_MIP_DIR = {
    'max': mip.MAXIMIZE,
    'min': mip.MINIMIZE
}

_MIP_DIR_TO_DIR = dict(
    [(val, key) for key, val in six.iteritems(_DIR_TO_MIP_DIR)]
)

# Needs to be used for to_bound during _initialize_model_from_problem
INFINITY = 1.7976931348623157e+308

def to_float(number, is_lb=True):
    """Converts None type and sympy.core.numbers.Float to float."""
    if number is not None:
        return float(number)
    if is_lb:
        return -INFINITY
    return INFINITY

def to_bound(number):
    """Convert float with infs to None."""
    if abs(number) == INFINITY:
        return None
    return number

def to_symbolic_expr(coeffs):
    """Converts coeffs dict to symbolic expression."""
    return symbolics.add([var * coef for var, coef in coeffs.items()
                          if not isclose(0, to_float(coef))])

@six.add_metaclass(inheritdocstring)
class Variable(interface.Variable):
    def __init__(self, name, *args, **kwargs):
        super(Variable, self).__init__(name, **kwargs)

    @interface.Variable.lb.setter
    def lb(self, value):
        interface.Variable.lb.fset(self, value)
        if self.problem is not None:
            self.problem._update_var_lb(self.name, value)

    @interface.Variable.ub.setter
    def ub(self, value):
        interface.Variable.ub.fset(self, value)
        if self.problem is not None:
            self.problem._update_var_ub(self.name, value)

    def set_bounds(self, lb, ub):
        super(Variable, self).set_bounds(lb, ub)
        if self.problem is not None:
            self.problem._update_var_lb(self.name, lb)
            self.problem._update_var_ub(self.name, ub)

    @interface.Variable.type.setter
    def type(self, value):
        if not value in _VTYPE_TO_MIP_VTYPE:
            raise ValueError(
                'COIN-OR CBC cannot handle variables of type %s. ' % value +
                'The following variable types are available:\n' +
                ' '.join(_VTYPE_TO_MIP_VTYPE.keys()))

        if self.problem is not None:
            self.problem._update_var_type(self.name, _VTYPE_TO_MIP_VTYPE[value])
        interface.Variable.type.fset(self, value)

        if value == 'integer':
            if self.lb is not None:
                self.lb = ceil(self.lb)
            if self.ub is not None:
                self.ub = floor(self.ub)
        elif value == 'binary':
            self.lb, self.ub = 0, 1

    @property
    def primal(self):
        if self.problem is None:
            return None
        return self.problem._var_primal(self.name)

    @property
    def dual(self):
        if self.type != 'continuous':
            raise ValueError('Dual is only available for continuous variables')
        if self.problem is None:
            return None
        return self.problem._var_dual(self.name)

    @interface.Variable.name.setter
    def name(self, value):
        super(Variable, Variable).name.fset(self, value)
        if getattr(self, 'problem', None) is not None:
            raise Exception('COIN-OR CBC doesn\'t support variable name change')


@six.add_metaclass(inheritdocstring)
class Constraint(interface.Constraint):
    _INDICATOR_CONSTRAINT_SUPPORT = False

    def __init__(self, expression, sloppy=False, *args, **kwargs):
        self._changed_expression = {}
        super(Constraint, self).__init__(expression, sloppy=sloppy, *args, **kwargs)

    def constraint_name(self, is_lb):
        if is_lb:
            return 'c_' + self.name + '_lower'
        return 'c_' + self.name + '_upper'

    @property
    def primal(self):
        if getattr(self, 'problem', None) is not None \
           and self.problem.status == interface.OPTIMAL:
            if self.lb is not None:
                return self.problem._constr_primal(self, True)
            if self.ub is not None:
                return self.problem._constr_primal(self, False)
        return None

    @property
    def dual(self):
        if getattr(self, 'problem', None) is None:
            return None
        if self.lb is not None:
            return self.problem._constr_dual(self, True)
        if self.ub is not None:
            return self.problem._constr_dual(self, False)
        return None

    def _update_bound(self, new, old, is_lb):
        """Updates associated mip model with new constraint bounds."""
        if getattr(self, 'problem', None) is None:
            return

        if old is None and new is not None:
            self.problem._add_mip_constraint(self, is_lb)
        elif new is None and old is not None:
            self.problem._remove_mip_constraint(self, is_lb)
        elif new is not None and old is not None:
            self.problem._update_constraint_bound(self, is_lb)

    @interface.Constraint.lb.setter
    def lb(self, value):
        self._check_valid_lower_bound(value)
        self._lb, old_lb = value, getattr(self, '_lb', None)
        self._update_bound(value, old_lb, True)

    @interface.Constraint.ub.setter
    def ub(self, value):
        self._check_valid_upper_bound(value)
        self._ub, old_ub = value, getattr(self, '_ub', None)
        self._update_bound(value, old_ub, False)

    @interface.Constraint.name.setter
    def name(self, value):
        super(Constraint, Constraint).name.fset(self, value)
        if getattr(self, 'problem', None) is not None:
            raise Exception('COIN-OR CBC doesn\'t support constraint name change')

    def _get_expression(self):
        if (self.problem is not None and self._changed_expression and
            len(self.problem._variables) > 0):

            coeffs = self._expression.as_coefficients_dict()

            new_vars = set(self._changed_expression) - set(coeffs)

            self._expression += symbolics.add([var * self._changed_expression[var]
                                              for var in new_vars])

            # Substitute var in expression with var * coef / old_coef
            updates = {var: var * coef / coeffs[var]
                       for var, coef in self._changed_expression.items()
                       if var not in new_vars}

            self._expression = self._expression.subs(updates)
            self._changed_expression = {}

        return self._expression

    def _get_mip_constr_expr(self):
        """Returns mip coefficient dictionary."""
        mip_constr = self.problem.problem.constr_by_name
        constr = mip_constr(self.constraint_name(is_lb=False))
        sign = 1
        if constr is None:
            constr = mip_constr(self.constraint_name(is_lb=True))
            sign = -1
        return (sign * constr.expr).expr

    def set_linear_coefficients(self, coefficients):
        if self.problem is None:
            raise Exception('Can\'t change coefficients if constraint is not associated with a model.')

        self.problem.update()

        mip_var = self.problem.problem.var_by_name

        expr = self._get_mip_constr_expr()
        names = set(var.name for var in coefficients)

        constr = mip.xsum(mip_var('v_' + var.name) * coef
                       for var, coef in coefficients.items()) \
               + mip.xsum(var * coef for var, coef in expr.items()
                       if var.name[2:] not in names)

        self.problem._remove_mip_constraint(self, True)
        self.problem._remove_mip_constraint(self, False)
        self.problem._add_mip_constraint(self, True, constr)
        self.problem._add_mip_constraint(self, False, constr)

        self._changed_expression.update(coefficients)

    def get_linear_coefficients(self, variables):
        if self.problem is None:
            raise Exception('Can\'t get coefficients from solver if constraint is not in a model')

        self.problem.update()
        mip_var = self.problem.problem.var_by_name
        expr = self._get_mip_constr_expr()
        return {v: expr.get(mip_var('v_' + v.name), 0) for v in variables}


@six.add_metaclass(inheritdocstring)
class Objective(interface.Objective):
    def __init__(self, expression, sloppy=False, **kwargs):
        self._changed_expression = {}
        super(Objective, self).__init__(expression, sloppy=sloppy, **kwargs)
        if not (sloppy or self.is_Linear):
            raise ValueError(
                'COIN-OR CBC only supports linear objectives. %s is not linear.' % self)

    @property
    def value(self):
        if getattr(self, 'problem', None) is None:
            return None
        return self.problem.problem.objective_value

    @interface.Objective.direction.setter
    def direction(self, value):
        super(Objective, self.__class__).direction.fset(self, value)
        if getattr(self, 'problem', None) is not None:
            self.problem.problem.sense = _DIR_TO_MIP_DIR[value]

    def _get_expression(self):
        if (self.problem is not None and self._changed_expression and
            len(self.problem._variables) > 0):

            coeffs = self._expression.as_coefficients_dict()

            new_vars = set(self._changed_expression) - set(coeffs)

            self._expression += symbolics.add([var * self._changed_expression[var]
                                              for var in new_vars])

            # Substitute var in expression with var * coef / old_coef
            updates = {var: var * coef / coeffs[var]
                       for var, coef in self._changed_expression.items()
                       if var not in new_vars}

            self._expression = self._expression.subs(updates)
            self._changed_expression = {}

        return self._expression

    def set_linear_coefficients(self, coefficients):
        if self.problem is None:
            raise Exception('Can\'t change coefficients if objective is not associated with a model.')

        self.problem.update()
        model = self.problem.problem
        expr = model.objective.expr
        names = set(var.name for var in coefficients)

        obj = mip.xsum(model.var_by_name('v_' + var.name) * coef
                       for var, coef in coefficients.items()) \
            + mip.xsum(var * coef for var, coef in expr.items()
                       if var.name[2:] not in names) \
            + model.objective.const

        # TODO: why does this not work? It would likely be faster
        # obj_update = mip.xsum(model.var_by_name('v_' + var.name) * coef
        #                       for var, coef in coefficients.items()) \
        #            - mip.xsum(-var * coef for var, coef in expr.items()
        #                       if var.name[2:] in names)
        # model.objective.add_expr(obj_update)

        self._changed_expression.update(coefficients)
        model.objective = obj

    def get_linear_coefficients(self, variables):
        if self.problem is None:
            raise Exception('Can\'t get coefficients from solver if objective is not in a model')

        self.problem.update()

        coeffs = self.problem.problem.objective.expr
        mip_var = self.problem.problem.var_by_name
        return {v: coeffs.get(mip_var('v_' + v.name), 0) for v in variables}


@six.add_metaclass(inheritdocstring)
class Configuration(interface.MathematicalProgrammingConfiguration):
    def __init__(self, verbosity=0, timeout=None, presolve='auto',
                 max_nodes=None, max_solutions=None, relax=False,
                 emphasis=0, cuts=-1, threads=0, *args, **kwargs):
        super(Configuration, self).__init__(*args, **kwargs)

        self.verbosity = verbosity

        # Enables/disables pre-processing. Pre-processing tries to improve your MIP
        # formulation. -1 means automatic, 0 means off and 1 means on.
        self.presolve = presolve

        # Time limit in seconds for search
        self.timeout = timeout

        # Maximum number of nodes to be explored in the search tree
        self.max_nodes = max_nodes

        # Solution limit, search will be stopped when max_solutions are found
        self.max_solutions = max_solutions

        # If true only the linear programming relaxation will be solved, i.e.
        # integrality constraints will be temporarily discarded. Changes the
        # type of all integer and binary variables to continuous. Bounds are preserved.
        self.relax = relax

        # 0. default setting: tries to balance between the search of improved feasible
        #    solutions and improvedlower bounds
        # 1. feasibility: focus on finding improved feasible solutions in the first
        #    moments of the search process,activates heuristics
        # 2. optimality: activates procedures that produce improved lower bounds,
        #    focusing in pruning thesearch tree even if the production of the first
        #    feasible solutions is delayed
        self.emphasis = emphasis

        # Controls the generation of cutting planes, -1 means automatic,
        # 0 disables completely, 1 (de-fault) generates cutting planes in
        # a moderate way, 2 generates cutting planes aggressively and 3
        # generates even more cutting planes. Cutting planes usually improve
        # the LP relaxation bound but also make the solution time of the LP
        # relaxation larger, so the overall effect is hardto predict and
        # experimenting different values for this parameter may be beneficial.
        self.cuts = cuts

        # Number of threads to be used when solving the problem. 0 uses solver
        # default configuration, -1 uses the number of available processing cores
        # and >= 1 uses the specified number of threads. An increased number of
        # threads may improve the solution time but also increases the memory consumption
        self.threads = threads

        if 'tolerances' in kwargs:
            for key, val in six.iteritems(kwargs['tolerances']):
                if key in self._tolerance_functions():
                    setattr(self.tolerances, key, val)

    @property
    def verbosity(self):
        return self._verbosity

    @verbosity.setter
    def verbosity(self, value):
        self._verbosity = value
        if getattr(self, 'problem', None) is not None:
            self.problem.problem.verbose = int(value > 1)

    @property
    def presolve(self):
        return self._presolve

    @presolve.setter
    def presolve(self, value):
        self._presolve = value
        if getattr(self, 'problem', None) is not None:
            value = -1 if value == 'auto' else int(value)
            self.problem.problem.preprocess = value

    @property
    def timeout(self):
        return self._timeout

    @timeout.setter
    def timeout(self, value):
        self._timeout = value
        if getattr(self, 'problem', None) is not None:
            if value is not None:
                self.problem.problem.max_seconds = value

    @property
    def max_nodes(self):
        return self._max_nodes

    @max_nodes.setter
    def max_nodes(self, value):
        self._max_nodes = value
        if getattr(self, 'problem', None) is not None:
            if value is not None:
                self.problem.problem.max_nodes = value

    @property
    def max_solutions(self):
        return self._max_solutions

    @max_solutions.setter
    def max_solutions(self, value):
        self._max_solutions = value
        if getattr(self, 'problem', None) is not None:
            if value is not None:
                self.problem.problem.max_solutions = value

    @property
    def relax(self):
        return self._relax

    @relax.setter
    def relax(self, value):
        self._relax = value

    @property
    def emphasis(self):
        return self._emphasis

    @emphasis.setter
    def emphasis(self, value):
        self._emphasis = value
        if getattr(self, 'problem', None) is not None:
            self.problem.problem.emphasis = value

    @property
    def cuts(self):
        return self._cuts

    @cuts.setter
    def cuts(self, value):
        self._cuts = value
        if getattr(self, 'problem', None) is not None:
            self.problem.problem.cuts = value

    @property
    def threads(self):
        return self._threads

    @threads.setter
    def threads(self, value):
        self._threads = value
        if getattr(self, 'problem', None) is not None:
            self.problem.problem.threads = value

    def __getstate__(self):
        return {
            'presolve': self.presolve,
            'timeout': self.timeout,
            'verbosity': self.verbosity,
            'max_nodes': self.max_nodes,
            'max_solutions': self.max_solutions,
            'relax': self.relax,
            'emphasis': self.emphasis,
            'cuts': self.cuts,
            'threads': self.threads,
            'tolerances': {
                'feasibility': self.tolerances.feasibility,
                'optimality': self.tolerances.optimality,
                'integrality': self.tolerances.integrality
                }
            }

    def __setstate__(self, state):
        for key, val in six.iteritems(state):
            if key != 'tolerances':
                setattr(self, key, val)
        # TODO: remove if check before final merge. Only here for backwards
        #       compatability for current pickle files stored in cobrapy
        if 'tolerances' in state:
            for key, val in six.iteritems(state['tolerances']):
                if key in self._tolerance_functions():
                    setattr(self.tolerances, key, val)

    def _get_feasibility(self):
        return self.problem.problem.infeas_tol

    def _set_feasibility(self, value):
        self.problem.problem.infeas_tol = value

    def _get_integrality(self):
        return self.problem.problem.integer_tol

    def _set_integrality(self, value):
        self.problem.problem.integer_tol = value

    def _get_optimality(self):
        return self.problem.problem.max_mip_gap_abs

    def _set_optimality(self, value):
        self.problem.problem.max_mip_gap_abs = value

    def _tolerance_functions(self):
        return {
            # 1e-6. Tightening this value can increase the numerical precision
            # but also probably increasethe running time. As floating point
            # computations always involve some loss of precision, values too
            # close to zero will likely render some models impossible to optimize.
            'feasibility': (
                self._get_feasibility,
                self._set_feasibility
            ),
            # Tolerance for the quality of the optimal solution, if a solution with
            # cost c and a lower bound b are available and c - b < mip_gap_abs,
            # the search will be concluded
            'optimality': (
                self._get_optimality,
                self._set_optimality
            ),
            # Maximum distance to the nearest integer for a variable to be
            # considered with an integervalue. Default value: 1e-6. Tightening
            # this value can increase the numerical precision but also probably
            # increase the running time. As floating point computations always
            # involve someloss of precision, values too close to zero will likely
            # render some models impossible to optimize.
            'integrality': (
                self._get_integrality,
                self._set_integrality
            )
        }


@six.add_metaclass(inheritdocstring)
class Model(interface.Model):

    def _initialize_problem(self):
        self.problem = mip.Model(solver_name=mip.CBC)

    def _initialize_model_from_problem(self, problem):
        if not isinstance(problem, mip.Model):
            raise TypeError('Problem must be an instance of mip.Model, not ' + repr(type(problem)))

        # Set problem
        self.problem = problem

        self.variables.clear()
        # Set variables
        for v in problem.vars:
            self.variables.append(Variable(name=v.name[2:],
                                           lb=to_bound(v.lb),
                                           ub=to_bound(v.ub),
                                           type=_MIP_VTYPE_TO_VTYPE[v.var_type],
                                           problem=self))

        self.constraints.clear()
        # Set constraints
        for c in problem.constrs:
            name, suffix = c.name[2:-6], c.name[-6:]
            if name not in self.constraints:
                expr = sum(coef * self.variables[var.name[2:]]
                           for var, coef in c.expr.expr.items())
                if suffix == '_lower':
                    expr *= -1
                self.constraints.append(Constraint(expr, name=name, problem=self))

            if suffix == '_lower':
                self.constraints[name].lb = to_bound(-1*c.rhs)
            else:
                self.constraints[name].ub = to_bound(c.rhs)

        # Set objective
        terms = sum(coef * self.variables[var.name[2:]]
                    for var, coef in problem.objective.expr.items())
        self._objective = Objective(terms + problem.objective.const,
                              direction=_MIP_DIR_TO_DIR[problem.sense],
                              problem=self)

    def _var_primal(self, name):
        primal = self.problem.var_by_name('v_' + name).x
        return 0. if primal is None else primal

    def _var_dual(self, name):
        dual = self.problem.var_by_name('v_' + name).rc
        return 0. if dual is None else dual

    def _get_primal_values(self):
        if getattr(self, '_status', '') == 'optimal':
            return super(Model, self)._get_primal_values()
        return [0. for _ in self.variables]

    def _get_reduced_costs(self):
        if getattr(self, '_status', '') == 'optimal':
            return super(Model, self)._get_reduced_costs()
        return [0. for _ in self.variables]

    def _update_var_lb(self, name, lb):
        self.problem.var_by_name('v_' + name).lb = to_float(lb, True)

    def _update_var_ub(self, name, ub):
        self.problem.var_by_name('v_' + name).ub = to_float(ub, False)

    def _update_var_type(self, name, var_type):
        self.problem.var_by_name('v_' + name).var_type = var_type

    def _add_variables(self, variables):
        super(Model, self)._add_variables(variables)
        for var in variables:
            # TODO: may need to handle obj, column options
            self.problem.add_var(name='v_' + var.name,
                                 var_type=_VTYPE_TO_MIP_VTYPE[var.type],
                                 lb=to_float(var.lb, True),
                                 ub=to_float(var.ub, False))

    def _remove_variables(self, variables):
        # TODO: optimization for removing all variables?
        if self.objective is not None:
            self.objective._changed_expression.update({var: 0 for var in variables})
        mip_vars = []
        for var in variables:
            name = var.name
            del self._variables_to_constraints_mapping[name]
            var.problem = None
            del self._variables[name]
            mip_vars.append(self.problem.var_by_name('v_' + name))
        self.problem.remove(mip_vars)

    def _constr_primal(self, con, is_lb):
        slack = self.problem.constr_by_name(con.constraint_name(is_lb)).slack
        return con.lb + slack if is_lb else con.ub - slack

    def _constr_dual(self, con, is_lb):
        return self.problem.constr_by_name(con.constraint_name(is_lb)).pi

    def _get_constraint_values(self):
        if getattr(self, '_status', '') == 'optimal':
            return super(Model, self)._get_constraint_values()
        return [0. for _ in self.constraints]

    def _get_shadow_prices(self):
        if getattr(self, '_status', '') == 'optimal':
            return super(Model, self)._get_shadow_prices()
        return [0. for _ in self.constraints]

    def _update_constraint_bound(self, con, is_lb):
        name = con.constraint_name(is_lb)
        constr = self.problem.constr_by_name(name)
        constr.rhs = to_float(-1*con.lb if is_lb else con.ub)

    def _expr_to_mip_expr(self, expr):
        """Parses mip linear expression from expression."""
        if hasattr(expr, "expression") and symbolics.USE_SYMENGINE:
            expr._expression = expr.expression.expand()
        offset, coeffs, _ = parse_optimization_expression(expr)
        return offset + mip.xsum(to_float(coef) * self.problem.var_by_name('v_' + var.name)
                                 for var, coef in coeffs.items())

    def _remove_mip_constraint(self, con, is_lb):
        name = con.constraint_name(is_lb)
        constr = self.problem.constr_by_name(name)
        if constr is not None:
            self.problem.remove(constr)

    def _add_mip_constraint(self, con, is_lb=True, constr=None):
        # Optimization for precomputing the parsed constr expression
        if constr is None:
            constr = self._expr_to_mip_expr(con)

        # TODO: check if mip supports indicator_variable
        name = con.constraint_name(is_lb)
        if is_lb and con.lb is not None:
           self.problem.add_constr(-constr <= -con.lb, name)
        elif not is_lb and con.ub is not None:
            self.problem.add_constr(constr <= con.ub, name)

    def _add_constraints(self, constraints, sloppy=False):
        super(Model, self)._add_constraints(constraints, sloppy=sloppy)
        for con in constraints:
            constr = self._expr_to_mip_expr(con)
            # Attempt to add lb and ub constraints
            self._add_mip_constraint(con, True, constr)
            self._add_mip_constraint(con, False, constr)

    def _remove_constraints(self, constraints):
        not_removed = True
        if len(constraints) > 350:  # Need to figure out a good threshold here
            keys = map(lambda c: c.name, constraints)
            self._constraints = self._constraints.fromkeys(set(self._constraints.keys()).difference(set(keys)))
            not_removed = False

        cons = []
        for con in constraints:

            if not_removed:
                try:
                    del self._constraints[con.name]
                except KeyError:
                    raise LookupError("Constraint %s not in solver" % con)
                con.problem = None

            cl = self.problem.constr_by_name(con.constraint_name(True))
            cu = self.problem.constr_by_name(con.constraint_name(False))
            if cl is not None:
                cons.append(cl)
            if cu is not None:
                cons.append(cu)

        self.problem.remove(cons)

    def _optimize(self):
        if self.configuration.relax:
            self.problem.relax()
            self._initialize_model_from_problem(self.problem)

        status = self.problem.optimize()
        # TODO: make more robust. See glpk_interface.py
        return _MIP_STATUS_TO_STATUS[status]

    @interface.Model.objective.setter
    def objective(self, value):
        super(Model, Model).objective.fset(self, value)
        self.update()
        self.problem.objective = self._expr_to_mip_expr(value)
        self.problem.sense = _DIR_TO_MIP_DIR[value.direction]
        value.problem = self

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        return result

    def to_lp(self):
        self.update()

        lp_form = ('Maximize' if self.problem.sense == mip.MAXIMIZE else 'Minimize') + '\n'
        for i, (var, coef) in enumerate(self.problem.objective.expr.items()):
            if i == 0:
                lp_form += f'OBJROW: {coef} {var.name}'
            else:
                sign = '+' if coef > 0 else '-'
                lp_form += f' {sign} {abs(coef)} {var.name}'

        lp_form += '\nSubject To\n'
        for con in self.problem.constrs:
            lp_form += f'{con}\n'

        lp_form += 'Bounds\n'
        for v in self.problem.vars:
            lp_form += f'{v.lb} <= {v} <= {v.ub}\n'
        lp_form += 'End\n'

        return lp_form

    @classmethod
    def from_lp(cls, lp_form):
        problem = mip.Model(solver_name=mip.CBC)
        with TemporaryFilename(suffix=".lp", content=lp_form) as tmp_file_name:
            problem.read(tmp_file_name)
        model = cls(problem=problem)
        return model
